class RequiredValuesNotProvidedDuringCreate(Exception):
    """Exception raised when the required values are not provided to the
    salesforce.create() function.

    Required Values:
        object_api: str -- The Object API name of the table to be inserted to in Salesforce.
        record_dataframe: pd.DataFrame() -- a DataFrame of the records to be inserted, including required fields are available
    """

    def __init__(self, message="Required parameters missing on create. An object_api and the record_dataframe must be provided"):
        self.message = message
        super().__init__(self.message)

class RequiredValuesNotProvidedDuringDelete(Exception):
    """Exception raised when the required values are not provided to the
    salesforce.delete() function.

    Required Values:
        object_api: str -- The Object API name of the table to be upserted to in Salesforce.
        record_dataframe: pd.DataFrame() -- a DataFrame of the records to be deleted. 
            The provided DataFrame must contain the "Id" column (Case sensitive)

    """

    def __init__(self, message="Required parameters missing on delete. \
        An object_api and the record_dataframe must be provided or the Id column missing in the provided dataframe."):
        self.message = message
        super().__init__(self.message)

class RequiredValuesNotProvidedDuringUpdate(Exception):
    """Exception raised when the required values are not provided to the
    salesforce.update() function.

    Required Values:
        object_api: str -- The Object API name of the table to be updated to in Salesforce.
        record_dataframe: pd.DataFrame() -- a DataFrame of the records to be updated. 
            The provided DataFrame must contain the "Id" column (Case sensitive)

    """

    def __init__(self, message="Required parameters missing on update. \
        An object_api and the record_dataframe must be provided or the Id column missing in the provided dataframe \
            or dataframe column names do not match a valid Salesforce field API name."):
        self.message = message
        super().__init__(self.message)

class RequiredValuesNotProvidedDuringUpsert(Exception):
    """Exception raised when the required values are not provided to the
    salesforce.upsert() function.

    Required Values:
        object_api: str -- The Object API name of the table to be upserted to in Salesforce.
        record_dataframe: pd.DataFrame() -- a DataFrame of the records to be upserted. 
            The provided DataFrame must contain the API name of an external ID column.

    """

    def __init__(self, message="Required parameters missing on upserted. \
        An object_api and the record_dataframe must be provided or and external ID column missing in the provided dataframe \
            or dataframe column names do not match a valid Salesforce field API name."):
        self.message = message
        super().__init__(self.message)

class RequiredFieldMissingDuringCreate(Exception):
    """Exception raised when the required fields are not provided to the
    salesforce.create() function.
    """
    def __init__(self, message="Required fields missing on create. Check your Salesforce datamodel to ensure all required fields are provided"):
        self.message = message
        super().__init__(self.message)




from sys import platform
import numpy as np
import pandas as pd
import os
from simple_salesforce import Salesforce
from simple_salesforce import SalesforceLogin
from salesforce_bulk import SalesforceBulk
from sfdclib import SfdcSession
from sfdclib import SfdcMetadataApi
from sfdclib import SfdcToolingApi
import re
import time
import json
from salesforce_bulk.util import IteratorBytesIO
import pandas as pd
import requests
import base64
from typing import List, Tuple, Optional, Union, Type
from collections import OrderedDict
import sys
sys.path.append("..")
try:
    from custom_exceptions.salesforce_exceptions import *
    from utilities.df_utils import *
    from utilities.async_utils import *
    from utilities.sf_query_processors import *
    from decorators import *
except:
    from salesforce.custom_exceptions.salesforce_exceptions import *
    from salesforce.utilities.df_utils import *
    from salesforce.utilities.async_utils import *
    from salesforce.utilities.sf_query_processors import *
    from salesforce.decorators import *
import asyncio
from functools import wraps, partial
from pandas import json_normalize




class Sf:
    _REGEX_PARSE_SOQL_FIELDS = "(?<=select)(.*?)(?<!, )(?=from)"
    _REGEX_PARSE_SOQL_OBJECT = "(?<!, from )(?<=from )\w*"
    
    def __init__(self) -> None:
        self.filename: str = None
        self.os_platform: str = platform
        self.credentials: pd.DataFrame = pd.DataFrame()
        self.sfUsername: str = None
        self.sfPassword: str = None
        self.sfOrgId: str = None
        self.isSandbox: bool = None
        self.session_id: str = None
        self.instance: str = None
        self.domain: str = None
        self.security_token: str = ''
        self.sf: Salesforce = None
        self.bulk: SalesforceBulk = None
        self.sfMeta: SfdcMetadataApi = None
        self.tooling: SfdcToolingApi = None
        self.api_version: str = 'v52.0'
        self.record_count: dict = {}
        self.record_count_caseinsensitive: dict = {}
        self.debug: bool = False
    
    def authenticate(self, sfUsername: Optional[str]=None, 
                                sfPassword: Optional[str]=None, 
                                sfOrgId: Optional[str]=None, 
                                isSandbox: Optional[bool]=None, 
                                session_id: Optional[str]=None, 
                                instance: Optional[str]=None, 
                                security_token: Optional[str] = None,
                                domain: Optional[str] = None,
                                if_return: Optional[bool] = False,
                                *args, **kwargs) -> Optional[dict]:
        """
        Authenticates Salesforce and retrieves the auth token.

        Dependencies:
            from simple_salesforce import Salesforce
            from simple_salesforce import SalesforceLogin
            from salesforce_bulk import SalesforceBulk
            from sfdclib import SfdcSession
            from sfdclib import SfdcMetadataApi
            from sfdclib import SfdcToolingApi
        """
        
        sfUsername = self.sfUsername if sfUsername is None else sfUsername
        sfPassword = self.sfPassword if sfPassword is None else sfPassword
        sfOrgId = self.sfOrgId if sfOrgId is None else sfOrgId
        isSandbox = self.isSandbox if isSandbox is None else isSandbox
        session_id = self.session_id if session_id is None else session_id
        instance = self.instance if instance is None else instance
        security_token = self.security_token if security_token is None else security_token
        domain = self.domain if domain is None else domain
        
        # If session ID already exists and instance URL is already populated,
        # reauthenticate using existing session ID
        if session_id is not None and instance is not None:
            sf = Salesforce(session_id = session_id, instance = instance)
            self.sf = sf
            self.instance = instance
            self.session_id = session_id
        
        # If username, password, org ID, and isSandbox flags are all provided,
        # authenticate using provided credentials
        elif sfUsername is not None and sfPassword is not None and sfOrgId is not None and isSandbox is not None:
            
            # SFDC Sandbox authentication
            if isSandbox:
                self.domain = 'test'
                sf = Salesforce(password=sfPassword, 
                                username=sfUsername, 
                                organizationId=sfOrgId, 
                                security_token = self.security_token,domain='test')
                session_id, instance = SalesforceLogin(
                username=sfUsername,
                password=sfPassword,
                security_token=self.security_token,
                domain= self.domain)
                self.session_id = session_id
                self.instance = instance
                self.sf = sf
                self.sfUsername = sfUsername
                self.sfPassword = sfPassword
                self.sfOrgId = sfOrgId
                self.isSandbox = isSandbox
                
            else:
                sf = Salesforce(password=sfPassword, 
                                username=sfUsername, 
                                organizationId=sfOrgId, 
                                security_token=self.security_token)
                session_id, instance = SalesforceLogin(
                username=sfUsername,
                password=sfPassword,
                security_token=self.security_token)
                self.session_id = session_id
                self.instance = instance
                self.sf = sf
                self.sfUsername = sfUsername
                self.sfPassword = sfPassword
                self.sfOrgId = sfOrgId
                self.isSandbox = isSandbox
                
        else:
            raise Exception('Either sfUsername, sfPassword, sfOrgId and isSandbox must be populated, OR session_id and instance must be populated.')

        # Alternative way to authenticate using SFDC Bulk API
        # bulk = SalesforceBulk(username=sfUsername, password=sfPassword, security_token='')
        bulk = SalesforceBulk(sessionId = self.session_id, host = self.instance)
        self.bulk = bulk
        # SFDC Metadata API
        sfMeta = SfdcSession(session_id=self.session_id, instance=".".join(self.instance.split(".")[:2]))
        self.sfMeta = sfMeta
        # Alternative way to authenticate using SFDC Metadata API
        # sfMeta = SfdcSession(username=sfUsername,password=sfPassword,token='',is_sandbox=isSandbox)
        sfMeta._api_version = "51.0"
        tooling = SfdcToolingApi(sfMeta)
        self.tooling = tooling
        
        self.api_version = 'v' + self.sf_api_call('/services/data')[-1]['version']
        
        for x in self.sf_api_call('/services/data/'+self.api_version+'/limits/recordCount')['sObjects']:
            self.record_count[x['name']] = x['count']
            self.record_count_caseinsensitive[x['name'].lower()] = x['count']
        
        if if_return:
            return {'sf':sf, 
                    'bulk':bulk, 
                    'sfMeta': sfMeta, 
                    'tooling':tooling, 
                    'session_id':session_id, 
                    'instance':instance, 
                    'sfMeta_is_connected':sfMeta.is_connected(), 
                    'bulk_api_sessionId':bulk.sessionId}

    ### ----------------------------------------------------------------------------------------------------
    ### Synchronous Data Functions
    ### ----------------------------------------------------------------------------------------------------

    def query(self, query: str, excludedFields: Optional[List] = [], *args) -> pd.DataFrame:
        """
        Using SFDC SOQL Syntax, and allowing for Relationships and group bys. 
        
        Arguments:
            query (str): A Standard SFDC SOQL Query allowing for relationships (Owner.Name)
                Asterisks(*) represents all queryable fields and can be used in conjunction
                with other relationship fields. 
                i.e. (Select *, Owner.Profile.Name, Owner.Name From Account)
        
        Returns:
            Pandas Dataframe Object.
        
        Raises:
            KeyError: Typically raised when 0 records exist for the object
                
            badfield: A self-correcting error that is raised when a field is unqueriable, i.e. Address Fields
            
            Exception: When relationship query contains more than 4 layers, an Exception is raised.
            i.e. Parent_Account_vod__r.Owner.Profile.LastModifiedBy.Name (<- a 5 layer deep relationship is not supported)
        
        Example of Usage:
            sf.query("Select *, Owner.Profile.Name From Account ORDER BY CreatedDate DESC LIMIT 100")
            
            return:
            A Pandas Dataframe of the last created 100 account records, with all queriable fields included in the query and a relationship field.
            
        """
        objectName = re.search(self._REGEX_PARSE_SOQL_OBJECT, query.lower()).group(0)
        successful = False
        
        extracted_object = pd.DataFrame()
        # replaces "*" in query with all fields on object
        sfSchema = getattr(self.sf, objectName).describe().get('fields')

        schemaDict = {}
        for x in sfSchema:
            schemaDict[x['name']] = x

        while not successful:
            try:
                results = []
                for field in schemaDict:
                    if (schemaDict[field]['type'] != 'location' and 
                        schemaDict[field]['type'] != 'address' and 
                        schemaDict[field]['name'] not in excludedFields):
                        results.append(field)
                final_query = query.replace("*", ", ".join(results))                
                query_response = self.sf.query_all(final_query)
                # if the object has 0 records in Salesforce, return empty dataframe
                if query_response['totalSize'] == 0:
                    fields_preparsed = re.search(self._REGEX_PARSE_SOQL_FIELDS, query,  re.IGNORECASE).group(0).split(",")
                    return pd.DataFrame(columns=fields_preparsed)
                else:
                    result = pd.DataFrame(query_response)['records']
                    for _ in result:
                        del _['attributes']
                    successful = True
            except KeyError:
                if self.debug:
                    print(objectName + ' skipped. (Potentially due to no records found.)')
                    
                return pd.DataFrame(columns=results)
            except Exception as badfield:
                field_exclusion = badfield.state_message[badfield.state_message.find("No such column '")+\
                    16:badfield.state_message.find("No such column '")+16+\
                        badfield.state_message[badfield.state_message.find("No such column '")+16:].find("'")]
                excludedFields.append(field_exclusion)
                print("Excluded unqueriable field: " + field_exclusion)
                if query.find("*") == -1:
                    raise Exception(f"Unqueriable field {field_exclusion} found in query.")
                else:
                    continue
            
            result = result.apply(lambda x: pd.Series(x)).copy()

            relationship_fields_preparsed = re.search(self._REGEX_PARSE_SOQL_FIELDS, query,  re.IGNORECASE).group(0).split(",")
            # relationship_fields_prepared Example:
            # ['*',
            #  ' Parent_Account_vod__r.Owner.Profile.Name',
            #  ' Child_Account_vod__r.Owner.Profile.Name',
            #  'Parent_Account_vod__r.Owner.Profile.Id ']

            relational_fields = [{x.strip(): x.strip().split(".")} for x in relationship_fields_preparsed if "." in x]
            # relationship_fields Example:
            # [{'Parent_Account_vod__r.Owner.Profile.Name': ['Parent_Account_vod__r',
            #    'Owner',
            #    'Profile',
            #    'Name']},
            #  {'Child_Account_vod__r.Owner.Profile.Name': ['Child_Account_vod__r',
            #    'Owner',
            #    'Profile',
            #    'Name']},
            #  {'Parent_Account_vod__r.Owner.Profile.Id': ['Parent_Account_vod__r',
            #    'Owner',
            #    'Profile',
            #    'Id']}]

            columns_to_remove = set()
            for x in relational_fields:
                if len(list(x.values())[0]) > 4:
                    raise Exception("Too Many Relationship Levels. The Query you have entered contains more than 4 levels deep and is not supported.")
                elif len(list(x.values())[0]) == 4:
                    result[list(x.keys())[0]] = result[list(x.keys())[0].split(".")[0]].\
                        apply(lambda z: z[list(x.values())[0][1]][list(x.values())[0][2]][list(x.values())[0][3]])
                    columns_to_remove.add(list(x.values())[0][0])
                elif len(list(x.values())[0]) == 3:
                    result[list(x.keys())[0]] = result[list(x.keys())[0].split(".")[0]].\
                        apply(lambda z: z[list(x.values())[0][1]][list(x.values())[0][2]])
                    columns_to_remove.add(list(x.values())[0][0])
                elif len(list(x.values())[0]) == 2:
                    result[list(x.keys())[0]] = result[list(x.keys())[0].split(".")[0]].\
                        apply(lambda z: z[list(x.values())[0][1]])
                    columns_to_remove.add(list(x.values())[0][0])
            result.drop(columns_to_remove, axis=1, inplace=True)
            result.replace(np.nan, None, inplace=True)        

                
        return result

    def create(self, object_api: str=None, record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
        """
        Function creates records within the Salesforce instance using bulk api.

        Arguments:
        object_api: str: API name of the Salesforce Object for which the records are to be created.

        record_dataframe: pd.DataFrame: A pandas dataframe object containing all the required fields for a record.

        Returns:
        result: pd.DataFrame: A pandas dataframe with the following columns:
            success: boolean: indicates whether the create request was successful
            created: boolean: this value should always be False in a create request
            id: object[str]: this value typically is None in a create request
            statusCode: object[str] - Optional: the error code of the create operation if failed
            message: object[str] - Optional: the error message of the create operation if failed
            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.

        Raises:
            RequiredValuesNotProvidedDuringUpdate
        """
        if (object_api is None) or (len(record_dataframe) == 0):
            raise RequiredValuesNotProvidedDuringCreate()
        else:
            result = getattr(self.sf.bulk, object_api).insert(record_dataframe.to_dict('records'), batch_size=10000, use_serial=False)
            result = pd.DataFrame(result)
            result = unpack_column(result, "errors")

        return pd.DataFrame(result.rename(columns={'id': 'Id'}))

    def delete(self, object_api: str=None, record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
        """
        Function deletes records within the Salesforce instance using bulk api.

        Arguments:
        object_api: str: API name of the Salesforce Object for which the records are to be deleted from.

        record_dataframe: pd.DataFrame: A pandas dataframe object containing at least 1 column with an Id column (case sensitive)

        Returns:
        result: pd.DataFrame: A pandas dataframe with the following columns:
            success: boolean: indicates whether the delete request was successful
            created: boolean: this value should always be False in a delete request
            id: object[str]: this value typically is None in a delete request
            statusCode: object[str] - Optional: the error code of the delete operation if failed
            message: object[str] - Optional: the error message of the delete operation if failed
            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.

        Raises:
            RequiredValuesNotProvidedDuringDelete
        """
        if (object_api is None) or (len(record_dataframe) == 0) or record_dataframe.columns.__contains__('Id') == False:
            raise RequiredValuesNotProvidedDuringDelete()
        else:
            result = getattr(self.sf.bulk, object_api).delete(record_dataframe['Id'].to_frame().to_dict('records'))
            result = pd.DataFrame(result)
            result = unpack_column(result, "errors")

        return pd.DataFrame(result.rename(columns={'id': 'Id'}))
    
    def update(self, object_api: str=None, record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
        """
        Function updates records within the Salesforce instance using bulk api.

        Arguments:
        object_api: str: API name of the Salesforce Object for which the records are to be updated.

        record_dataframe: pd.DataFrame: A pandas dataframe object containing at least 1 column with an Id column (case sensitive)
            and additional columns matching the Salesforce field api names.

        Returns:
        result: pd.DataFrame: A pandas dataframe with the following columns:
            success: boolean: indicates whether the update request was successful
            created: boolean: this value should always be False in a update request
            id: object[str]: this value typically is None in a update request
            statusCode: object[str] - Optional: the error code of the update operation if failed
            message: object[str] - Optional: the error message of the update operation if failed
            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.

        Raises:
            RequiredValuesNotProvidedDuringUpdate
        """

        field_metadata = self.field_describe([object_api],attributes=['name','type','updateable','compoundFieldName'])

        if ((object_api is None) or 
        (len(record_dataframe) == 0)):
            raise RequiredValuesNotProvidedDuringUpdate(message="Object API and Record Dataframe are required, ensure your record dataframe has at least 1 row and 1 column with an Id column")
        elif record_dataframe.columns.__contains__('Id') == False:
            raise RequiredValuesNotProvidedDuringUpdate(message="Record Dataframe must have an Id column")
        elif ~record_dataframe.columns.isin(field_metadata[object_api]).all():
            # checks whether any of the columns in the dataframe passed in are not valid field API names
            raise RequiredValuesNotProvidedDuringUpdate(message="One of the columns in the record dataframe is not a valid field for the object")
        elif record_dataframe.columns.isin(field_metadata[(field_metadata['Updateable'] == False) & (field_metadata[object_api] != "Id")][object_api]).any():
            # checks whether any of the columns in the dataframe passed in are not updateable
            non_updatable_fields = list(set(record_dataframe.columns) - set(field_metadata[(field_metadata['Updateable'] == True) | (field_metadata[object_api] == "Id")][object_api]))
            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are not updatable: {', '.join(non_updatable_fields)}")
        elif record_dataframe.columns.isin(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')][object_api]).any():
            # checks whether any of the columns in the dataframe passed in are compound fields
            compound_fields = list(set(record_dataframe.columns) & set(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')][object_api]))
            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are compound fields, which are not updatable: {', '.join(compound_fields)}")
        # checks whether an reference (lookup/master-detail) field is included in the source dataframe
        #  or ~record_dataframe.columns.isin(field_metadata[field_metadata['Type'] == 'reference'][object_api]).any()
            
        else:
            record_dataframe.replace(np.nan, None, inplace=True)
            result = getattr(self.sf.bulk, object_api).update(record_dataframe.to_dict('records'), batch_size=10000, use_serial=False)
            result = pd.DataFrame(result)
            result = unpack_column(result, "errors")

        return pd.DataFrame(result.rename(columns={'id': 'Id'}))

    def upsert(self, object_api: str=None, external_id_field_api: str="", record_dataframe: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
        """
        Function upserts records within the Salesforce instance using bulk api.

        Arguments:
        object_api: str: API name of the Salesforce Object for which the records are to be upserted into.

        external_id_field_api: str: API name of the external ID field used for the upsert operation.

        record_dataframe: pd.DataFrame: A pandas dataframe object containing the External column (case sensitive)
            and additional columns matching the Salesforce field api names.

        Returns:
        result: pd.DataFrame: A pandas dataframe with the following columns:
            success: boolean: indicates whether the upsert request was successful
            created: boolean: this value should always be False in a upsert request
            id: object[str]: this value typically is None in a upsert request
            statusCode: object[str] - Optional: the error code of the upsert operation if failed
            message: object[str] - Optional: the error message of the upsert operation if failed
            fields: object[List[str]] - Optional: the fields for which the error code and message applies to.

        Raises:
            RequiredValuesNotProvidedDuringUpsert
            SalesforceMalformedRequest
        """

        field_metadata = self.field_describe([object_api], ['name', 'type', 'length','externalId','updateable','compoundFieldName'])
            
            
        if ((object_api is None) or 
        (len(record_dataframe) == 0)):
            raise RequiredValuesNotProvidedDuringUpdate(message="Object API and Record Dataframe are required, ensure your record dataframe has at least 1 row and 1 column with an Id column")
        elif ~record_dataframe.columns.isin(field_metadata[object_api]).all():
            # checks whether any of the columns in the dataframe passed in are not valid field API names
            raise RequiredValuesNotProvidedDuringUpdate(message="One of the columns in the record dataframe is not a valid field for the object")
        elif external_id_field_api == "":
            raise RequiredValuesNotProvidedDuringUpsert(message="External ID field API is required")
        elif record_dataframe.columns.isin(field_metadata[(field_metadata['Updateable'] == False) & (field_metadata[object_api] != "Id")][object_api]).any():
            # checks whether any of the columns in the dataframe passed in are not updateable
            non_updatable_fields = list(set(record_dataframe.columns) - set(field_metadata[(field_metadata['Updateable'] == True) | (field_metadata[object_api] == "Id")][object_api]))
            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are not updatable: {', '.join(non_updatable_fields)}")
        elif record_dataframe.columns.isin(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')][object_api]).any():
            # checks whether any of the columns in the dataframe passed in are compound fields
            compound_fields = list(set(record_dataframe.columns) & set(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')][object_api]))
            raise RequiredValuesNotProvidedDuringUpdate(message=f"The following fields are compound fields, which are not updatable: {', '.join(compound_fields)}")
        else:
            record_dataframe.replace(np.nan, None, inplace=True)
            result = getattr(self.sf.bulk, object_api).upsert(record_dataframe.to_dict('records'), external_id_field_api, batch_size=10000, use_serial=False)
            result = pd.DataFrame(result)
            result = unpack_column(result, "errors")

        return pd.DataFrame(result)

    def extract_bulk(self, og_query: str, 
                        excludedFields: Optional[List] = []) -> pd.DataFrame:
        """
        Uses a standard SOQL query to extract Salesforce Data and outputs a pandas dataframe
        
        Dependencies:
            import re
            import time
            import json
            from salesforce_bulk.util import IteratorBytesIO
            import pandas as pd
        
        """
        objectName = re.search(self._REGEX_PARSE_SOQL_OBJECT, og_query.lower()).group(0)
        successful = False
        
        extracted_object = pd.DataFrame()
        # replaces "*" in query with all fields on object
        sfSchema = getattr(self.sf, objectName).describe().get('fields')

        schemaDict = {}
        for x in sfSchema:
            schemaDict[x['name']] = x

        while not successful:
            try:
                results = []
                for field in schemaDict:
                    if (schemaDict[field]['type'] != 'location' and 
                        schemaDict[field]['type'] != 'address' and 
                        schemaDict[field]['name'] not in excludedFields):
                        results.append(field)
                query = og_query.replace("*", ", ".join(results))
                
#                 # if the object has 0 records in Salesforce, return empty dataframe
#                 if objectName not in self.record_count_caseinsensitive.keys():
#                     return pd.DataFrame(columns=results)
                    
                job = self.bulk.create_query_job(objectName, contentType='JSON')
                batch = self.bulk.query(job, query)
                while not self.bulk.is_batch_done(batch):
                    time.sleep(1)
                sfdf = pd.DataFrame()
                for result in self.bulk.get_all_results_for_query_batch(batch):
                    result = json.load(IteratorBytesIO(result))
                    sfdf = sfdf.append(pd.DataFrame(result))

                # drops attributes column in dataframe
                sfdf.drop(columns="attributes", inplace = True)

                # formats all datetime to the proper formatting
                for column in sfdf:
                    if schemaDict[column]['type'] == 'datetime':
                        sfdf[column] = pd.to_datetime(sfdf[column], unit='ms')
                    # if the column has a 'scale' or salesforce's decimal places, then turn the column into an int
                    elif schemaDict[column]['type'] == 'double' and schemaDict[column]['scale'] == 0:
                        sfdf[column] = pd.to_numeric(sfdf[column], downcast='integer')

            #                         pd.to_datetime(sfdf[column], unit = 's')
            #                         sfdf[column].apply(lambda x : datetime.fromtimestamp(int(x), tz).isoformat())
            #                 sfdf.convert_dtypes()
                # converts the output of the bulk query to text so that the unix timestamp displays property, and fills and empty values with the '' string.
                sfdf = sfdf.fillna('').astype(str)
                successful = True
            except KeyError:
                if self.debug:
                    print(objectName + ' skipped. (Potentially due to no records found.)')
                    
                return pd.DataFrame(columns=results)
            except Exception as badfield:
                field_exclusion = badfield.state_message[badfield.state_message.find("No such column '")+\
                    16:badfield.state_message.find("No such column '")+16+\
                        badfield.state_message[badfield.state_message.find("No such column '")+16:].find("'")]
                excludedFields.append(field_exclusion)
                print("Excluded unqueriable field: " + field_exclusion)
                if og_query.find("*") == -1:
                    raise Exception(f"Unqueriable field {field_exclusion} found in query.")
                else:
                    continue
        if self.debug:
            print("Extracted " + objectName + " successfully!")
        return sfdf

    ### ----------------------------------------------------------------------------------------------------
    ### Asynchronous Data Functions
    ### ----------------------------------------------------------------------------------------------------

    async def async_query(self, query, excludedFields: Optional[List]= []) -> pd.DataFrame:
        async_query = async_wrap(self.query)
        return await async_query(query, excludedFields)

    async def async_upsert(self, object_api, record_dataframe, external_id_field_api, batchsize=2000, *args, **kwargs):
        async_upsert = async_wrap(self.upsert)
        batches = len(record_dataframe) / batchsize
        result_list = await asyncio.gather(*[async_upsert(object_api= object_api, record_dataframe=batch, external_id_field_api=external_id_field_api) for batch in np.array_split(record_dataframe, batches)])
        return pd.concat(result_list).reset_index(drop=True)

    async def async_update(self, object_api, record_dataframe, batchsize=2000, *args, **kwargs):
        async_update = async_wrap(self.update)
        batches = len(record_dataframe) / batchsize
        result_list = await asyncio.gather(*[async_update(object_api= object_api, record_dataframe=batch) for batch in np.array_split(record_dataframe, batches)])
        return pd.concat(result_list).reset_index(drop=True)

    async def async_create(self, object_api, record_dataframe, batchsize=2000, *args, **kwargs):
        async_create = async_wrap(self.create)
        batches = len(record_dataframe) / batchsize
        result_list = await asyncio.gather(*[async_create(object_api= object_api, record_dataframe=batch) for batch in np.array_split(record_dataframe, batches)])
        return pd.concat(result_list).reset_index(drop=True)

    async def async_delete(self, object_api, record_dataframe, batchsize=2000, *args, **kwargs):
        async_delete = async_wrap(self.delete)
        batches = len(record_dataframe) / batchsize
        result_list = await asyncio.gather(*[async_delete(object_api= object_api, record_dataframe=batch) for batch in np.array_split(record_dataframe, batches)])
        return pd.concat(result_list).reset_index(drop=True)

    ### ----------------------------------------------------------------------------------------------------
    ### Asynchronous Metadata Functions
    ### ----------------------------------------------------------------------------------------------------
    async def async_get_non_updatable_fields(self, object_name: str):
        """
        Returns a list of fields that cannot be updated.
        """
        async_field_describe =  async_wrap(self.field_describe)
        field_metadata = await async_field_describe([object_name], attributes=['name','updateable'])
        non_updatable_fields = list(field_metadata[field_metadata['Updateable'] == False][object_name])[1:]
        return non_updatable_fields

    async def async_get_compound_field_names(self, object_name: str, include_name_compound_fields: bool=False):
        """
        Returns a list of compound fields. By default "Name" compound field names are excluded.
        To return a list of compound fields including "Name" fields, set include_name_compound_fields to True.

        Compound field names are fields that are composed of multiple fields. These fields are not updatable directly via the API
        and may cause errors if updated directly.
        """
        async_field_describe =  async_wrap(self.field_describe)
        field_metadata = await async_field_describe([object_name], attributes=['name','compoundFieldName'])

        if include_name_compound_fields:
            compound_field_names = list(field_metadata[(~field_metadata['Compoundfieldname'].isnull())][object_name])
        elif include_name_compound_fields==False:
            compound_field_names = list(field_metadata[(~field_metadata['Compoundfieldname'].isnull()) & (field_metadata['Compoundfieldname'] != 'Name')][object_name])
        else:
            raise ValueError('include_name_compound_fields must be a boolean.')
        return compound_field_names
    
    
    async def get_picklist_values_by_object_record_type(self, object: str):
        object_record_type_ids = self.query(f"Select Id, Name From RecordType Where SobjectType = '{object}'").apply(lambda row: {"Id": row['Id'], "Name": row['Name']}, axis=1)
        if len(object_record_type_ids) == 0:
            object_record_type_ids = pd.Series([{"Id": "012000000000000AAA", "Name": "Master"}]).to_list()
        else:
            object_record_type_ids = object_record_type_ids.to_list()
            
        object_describe = self.sf_api_call(f"/services/data/{self.api_version}/ui-api/object-info/{object}")['fields']
        controlling_fields = []
        for field, values in object_describe.items():
            controlling_fields.append({"Field": field, "Controlling Field": "" if values['controllingFields'] == [] else "; ".join(values['controllingFields'])})
        controlling_fields = pd.DataFrame(controlling_fields)

        def record_type_picklist_retriever(self, object, record_type_id, record_type_name):
            result = []
            for key, value in self.sf_api_call(f"/services/data/{self.api_version}/ui-api/object-info/{object}/picklist-values/{record_type_id}")['picklistFieldValues'].items():
                result.append({"Field_API": key, "controllerValues": value['controllerValues'], "defaultValue":  value['defaultValue'], "values": value['values']})
            final_result_pd = json_normalize(result, ['values'], ['defaultValue', 'controllerValues', 'Field_API'])
            ## Converts the validFor numerical codes i.e. [0,2] to it's corresponding text values by looking it up on the controllerValues column i.e. {'Web': 0, 'Phone Inquiry': 1, 'Partner Reference': 2 ... }
            ## [0,2] -> ['Web', 'Partner Reference']
            final_result_pd['validFor'] = final_result_pd.apply(lambda row: "\n".join([{v: k for k, v in row['controllerValues'].items()}[value] for value in row['validFor']]), axis=1)
            final_result_pd.drop(columns=['controllerValues'], inplace=True)
            final_result_pd[['defaultValueLabel', 'defaultValueValue']] = json_normalize(final_result_pd['defaultValue'])[['label','value']]
            final_result_pd.drop(columns=['defaultValue'], inplace=True)
            final_result_pd = pd.merge(final_result_pd, controlling_fields, left_on='Field_API', right_on='Field', how='left').drop(columns=['Field']).copy()
            final_result_pd['RecordTypeName'] = record_type_name
            final_result_pd['Object'] = object
            return final_result_pd

        async_picklist_rt_retrieve = async_wrap(record_type_picklist_retriever)
        results = pd.concat(await asyncio.gather(*[async_picklist_rt_retrieve(self, object, record_type_id['Id'], record_type_id['Name']) for record_type_id in object_record_type_ids]))
        results = results[['Object', 'RecordTypeName', 'Field_API', 'Controlling Field', 'defaultValueLabel', 'defaultValueValue', 'label', 'value', 'validFor']]
        return results



    async def object_permission_bulk_check(self, objects: list, permissions: list ):
        async_object_permission_check = async_wrap(self.object_permission_check)
        result_list = await asyncio.gather(*[async_object_permission_check(object, permission ) for object in objects for permission in permissions])
        # for result in result_list:
        #     data = deep_merge_dictionaries(data, result)
        result_list = [item for sublist in result_list for item in sublist]
        return result_list

    async def field_permission_bulk_check(self, objects: list, ) -> dict[str, pd.DataFrame]:
        async_field_permission_check = async_wrap(self.field_permission_check)
        result_list = await asyncio.gather(*[async_field_permission_check( object ) for object in objects])
        # for result in result_list:
        #     data = deep_merge_dictionaries(data, result)
        profile_list = [item['profilePermissions'] for item in result_list]
        permissionSet_list = [item['permissionSetPermissions'] for item in result_list]
        result_dict = {'profilePermissions': pd.concat(profile_list), 'permissionSetPermissions': pd.concat(permissionSet_list)}
        
        return result_dict
    
    ### ----------------------------------------------------------------------------------------------------
    ### Synchronous Metadata Functions
    ### ----------------------------------------------------------------------------------------------------

    def object_describe(self, sobject_api_name: str, export_excel: bool=False):
        """

        """
        sfSchema = getattr(self.sf, sobject_api_name).describe()
        boolMetadata = {}
        nonetypeMetadata = {}
        strMetadata = {}
        orderedDictMetadata = pd.Series(dtype='object')
        childRelationshipsPD = pd.DataFrame()
        recordTypeInfosPD = pd.DataFrame()
        fieldsPD = pd.DataFrame()
        supportedScopesPD = pd.DataFrame()
        namedLayoutInfosPD = pd.DataFrame()
        actionOverridesPD = pd.DataFrame()

        for metadata in sfSchema:
            ## Processes all metadata that would return a boolean datatype
            if type(sfSchema[metadata]) is bool:
                boolMetadata[metadata] = sfSchema[metadata]

            ## Processes all metadata that would return a NoneType datatype
            elif type(sfSchema[metadata]) == type(None):
                nonetypeMetadata[metadata] = sfSchema[metadata]

            ## Processes all metadata that would return a string datatype
            elif type(sfSchema[metadata]) is str:
                strMetadata[metadata] = sfSchema[metadata]
            ## Processes all metadata that would return an OrderedDict data type
            elif type(sfSchema[metadata]) is OrderedDict and len(sfSchema[metadata]) != 0:
                orderedDictMetadata = pd.concat([orderedDictMetadata,pd.Series(sfSchema['urls'])])
            elif type(sfSchema[metadata]) is OrderedDict and len(sfSchema[metadata]) == 0:
                nonetypeMetadata[metadata] = "None"

            ## Processes all metadata that would return a list data type
            elif type(sfSchema[metadata]) is list and len(sfSchema[metadata]) != 0:
                if metadata == 'childRelationships':
                    childRelationshipsPD = pd.DataFrame(sfSchema['childRelationships'], index = pd.DataFrame(sfSchema['childRelationships'])['field'])
                    childRelationshipsPD = childRelationshipsPD.drop('field', axis=1).T
                if metadata == 'recordTypeInfos':
                    recordTypeInfosPD = pd.DataFrame(sfSchema['recordTypeInfos'], index = pd.DataFrame(sfSchema['recordTypeInfos'])['name'])
                    recordTypeInfosPD = recordTypeInfosPD.drop('name', axis = 1).T  
                if metadata == 'fields':
                    fieldsPD = pd.DataFrame(sfSchema['fields'], index = pd.DataFrame(sfSchema['fields'])['name'])
                    fieldsPD = fieldsPD.drop('name', axis = 1).T
                if metadata == 'supportedScopes':
                    supportedScopesPD = pd.DataFrame(sfSchema['supportedScopes'], index = pd.DataFrame(sfSchema['supportedScopes'])['name'])
                    supportedScopesPD = supportedScopesPD.drop('name', axis = 1).T
                if metadata == 'actionOverrides':
                    actionOverridesPD = pd.DataFrame(sfSchema['actionOverrides'], index = pd.DataFrame(sfSchema['actionOverrides'])['name'])
                    actionOverridesPD = actionOverridesPD.drop('name', axis = 1).T
                if metadata == 'namedLayoutInfos':
                    namedLayoutInfosPD = pd.DataFrame(sfSchema['namedLayoutInfos'], index = pd.DataFrame(sfSchema['namedLayoutInfos'])['name'])
                    namedLayoutInfosPD = namedLayoutInfosPD.drop('name', axis = 1).T
            ## Parses all empty list metadata
            elif type(sfSchema[metadata]) is list and len(sfSchema[metadata]) == 0:
                nonetypeMetadata[metadata] = "None"
            else:
                nonetypeMetadata[metadata] = "Unrecognized metadata type: " + metadata

        output = {'metadata': pd.concat([pd.Series(nonetypeMetadata),
                    pd.Series(boolMetadata),
                    pd.Series(strMetadata),
                    orderedDictMetadata]),
                    'childRelationships': childRelationshipsPD if len(childRelationshipsPD)!=0 else None,
                    'recordTypeInfos': recordTypeInfosPD if len(recordTypeInfosPD)!=0 else None,
                    'fields': fieldsPD if len(fieldsPD)!=0 else None,
                    'supportedScopes': supportedScopesPD if len(supportedScopesPD)!=0 else None, 
                    'namedLayoutInfos': namedLayoutInfosPD if len(namedLayoutInfosPD)!=0 else None,
                    'actionOverrides': actionOverridesPD if len(actionOverridesPD)!=0 else None}
        if export_excel:
            with pd.ExcelWriter(sobject_api_name + ' describe output.xlsx') as writer:  
                for key in output.keys():
                    try:
                        pd.DataFrame(output[key]).to_excel(writer, sheet_name=key)
                    except:
                        continue

        return output

    def field_describe(self, objects: List = ['Account','Address_vod__c','Child_Account_vod__c'], 
    attributes: List = ['name','type','length']) -> pd.DataFrame:
        """
        Returns a dataframe of the field metadata for the specified objects and attributes.

        Parameters
        ----------
        objects : List, optional
            A list of objects to get field metadata for. The default is ['Account','Address_vod__c','Child_Account_vod__c'].
        attributes : List, optional
            A list of attributes to get field metadata for. The default is ['name','type','length'].
            For a full list of attributes, see 
            https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_calls_describesobjects_describesobjectresult.htm
            A list of attributes can also be found by using the object_describe() method's field attribute.

        Returns
        -------
        pd.DataFrame
            A dataframe of the field metadata for the specified objects and attributes.
        
        """
        outputList = []
        columnNames = []
        for sObjectAPIName in objects:
            for attribute in attributes:
                outputList.append([field[attribute] for field in getattr(self.sf, sObjectAPIName).describe()['fields']])
                if attribute == "name":
                    columnNames.append(sObjectAPIName)
                else:
                    columnNames.append(attribute.title())
        field_describe = pd.DataFrame(outputList).transpose()
        field_describe.columns = columnNames
        return field_describe

    def picklist_dataframe_stacked(self,objects: List =['Account','Address_vod__c','Child_Account_vod__c']) -> pd.DataFrame:
        """
        TODO:
        
        Description of what it does
        
        Description of arguments and data types
        
        Description of return values and data types
        
        Description of Errors raised
        
        Extra Notes and Examples of Usage
        """
        output_df = pd.DataFrame()
        for object in objects:
            objectDescribe = getattr(self.sf, object).describe()
            processing_df = pd.DataFrame(pd.DataFrame([pd.Series(data = [picklist['value'] for picklist in field['picklistValues']], 
                                                                   name = object + "." + field["name"]) for field in objectDescribe['fields'] if field['type'] == 'picklist']).stack())
            processing_df.columns = ['Picklist API Value']
            processing_df['CRM Object and Field API'] = processing_df.index.get_level_values(0)
            processing_df[['CRM Object API','CRM Field API']] = processing_df['CRM Object and Field API'].str.split(".", expand = True)
            processing_df.reset_index(drop=True, inplace=True)
            output_df = output_df.append(processing_df)
        return output_df

    def picklist_dataframe(self,objects = ['Account','Address_vod__c','Child_Account_vod__c']) -> List:
        """
        TODO:
        
        Description of what it does
        
        Description of arguments and data types
        
        Description of return values and data types
        
        Description of Errors raised
        
        Extra Notes and Examples of Usage
        """
        referenceList = []
        for object in objects:
            objectDescribe = getattr(self.sf, object).describe()
            objectPicklistValues = pd.DataFrame(index=range(0,max(len(field["picklistValues"]) for field in objectDescribe['fields'] if field['type'] == 'picklist')))
            for x in [pd.Series(data = [picklist['value'] for picklist in field['picklistValues']], name = object + "." + field["name"]) for field in objectDescribe['fields'] if field['type'] == 'picklist']:
                objectPicklistValues.insert(0, str(x.name), x)
            referenceList.append(objectPicklistValues)
        return referenceList

    def record_type_retrieval(self, objectAPIName, fieldAPINames = ["Id","Name",'SobjectType']):
        """
        TODO:
        
        Description of what it does
        
        Description of arguments and data types
        
        Description of return values and data types
        
        Description of Errors raised
        
        Extra Notes and Examples of Usage
        """
        sfRTDF = pd.DataFrame(self.sf.query_all("SELECT "+ ",".join(fieldAPINames) + " from RecordType WHERE SobjectType = '" + objectAPIName + "'")['records'])
        sfRTDF.drop(columns="attributes", inplace = True)
        return sfRTDF

    def object_permission_check(self, object, permission):
        data = transform_sf_result_set_rec(self.sf.query_all(f"""SELECT Assignee.Id, Assignee.Name,Assignee.IsActive, Assignee.UserName, PermissionSet.Id, PermissionSet.isOwnedByProfile, PermissionSet.Profile.Name, PermissionSet.Label
            FROM PermissionSetAssignment
            WHERE PermissionSetId
            IN (SELECT ParentId
            FROM ObjectPermissions
            WHERE SObjectType IN ('{object}') AND
            ({permission} = true)) AND Assignee.IsActive = TRUE""")['records'])
        
        profile_violations = data[data['PermissionSet.IsOwnedByProfile']== True][['User.Username', 'Profile.Name']]
        permission_set_violations = data[data['PermissionSet.IsOwnedByProfile']== False][['User.Username', 'PermissionSet.Label']]
        result = []
        for user in profile_violations['User.Username'].unique():
            result.append({"UserName": user, 
                            "Object API Name" : object,
                            "Permission": permission,
                            "Profile": "; ".join(list(profile_violations[profile_violations['User.Username'] == user]['Profile.Name'].unique())),
                            "Permission Set": ""
                        
                        }
                        
                        )
            
        for user in permission_set_violations['User.Username'].unique():
            result.append({"UserName": user, 
                                                            "Object API Name" : object,
                                                            "Permission": permission,
                                                            "Profile": "",
                                                            "Permission Set": "; ".join(list(permission_set_violations[permission_set_violations['User.Username'] == user]['PermissionSet.Label'].unique()))})
        return result

    def field_permission_check(self, object) -> dict[str, pd.DataFrame]:
        fls_results = transform_sf_result_set_rec(self.sf.query_all(f"""
                                            SELECT Field,ParentId,PermissionsEdit,PermissionsRead,SobjectType FROM FieldPermissions
                                            WHERE SobjectType in ('{object}')
                                            """)['records'])

        permissionSetProfiles = transform_sf_result_set_rec(self.sf.query_all("""
                                                    SELECT Id,Description,IsOwnedByProfile,Name,ProfileId, Profile.Name, Profile.Description,Type FROM PermissionSet
                                                    """)['records'])

        profiles = permissionSetProfiles[permissionSetProfiles['PermissionSet.IsOwnedByProfile'] == True][['PermissionSet.ProfileId','PermissionSet.Id','Profile.Name','Profile.Description']].copy()
        permissionSets = permissionSetProfiles[permissionSetProfiles['PermissionSet.IsOwnedByProfile'] == False][['PermissionSet.Id','PermissionSet.Name','PermissionSet.Description']].copy()

        permissionSetPermissions = pd.merge(fls_results, permissionSets, left_on='FieldPermissions.ParentId', right_on='PermissionSet.Id', how='inner').drop(columns=['PermissionSet.Id'], axis=1)
        permissionSetPermissions.columns = ['Field API Name', 'Permission Set Id', 'FLS Edit', 'FLS Read', 'Object API Name', 'Permission Set Name', 'Permission Set Description']
        permissionSetPermissions = permissionSetPermissions[['Permission Set Id', 'Permission Set Name', 'Permission Set Description', 'Object API Name', 'Field API Name', 'FLS Edit', 'FLS Read']]
        permissionSetPermissions['Field API Name'] = permissionSetPermissions['Field API Name'].apply(lambda x: x.split('.')[-1])


        profilePermissions = pd.merge(fls_results, profiles, left_on='FieldPermissions.ParentId', right_on='PermissionSet.Id', how='inner').drop(columns=['PermissionSet.Id','FieldPermissions.ParentId'], axis=1)
        profilePermissions.columns = ['Field API Name', 'FLS Edit', 'FLS Read', 'Object API Name','Profile Id',  'Profile Name', 'Profile Description']
        profilePermissions = profilePermissions[['Profile Id', 'Profile Name', 'Profile Description', 'Object API Name', 'Field API Name', 'FLS Edit', 'FLS Read']]
        profilePermissions['Field API Name'] = profilePermissions['Field API Name'].apply(lambda x: x.split('.')[-1])
        
        return {'permissionSetPermissions': permissionSetPermissions, 'profilePermissions': profilePermissions}



    @serialze_zeep
    def metadata_read(self, metadata_type: str, fullName: str):
        return getattr(self.sf.mdapi, metadata_type).read(fullName)

    def metadata_delete(self, metadata_type: str, fullName: str):
        return getattr(self.sf.mdapi, metadata_type).delete(fullName)

    def metadata_update(self, metadata_type: str, parsed_metadata: dict):
        mdapi_object = getattr(self.sf.mdapi, metadata_type)()
        for key in parsed_metadata:
            setattr(mdapi_object, key, parsed_metadata[key])
        getattr(self.sf.mdapi, metadata_type).update(mdapi_object)
        return mdapi_object

    def metadata_create(self, metadata_type: str, parsed_metadata: dict):
        mdapi_object = getattr(self.sf.mdapi, metadata_type)()
        for key in parsed_metadata:
            setattr(mdapi_object, key, parsed_metadata[key])
        getattr(self.sf.mdapi, metadata_type).create(mdapi_object)
        return mdapi_object

    def metadata_list(self, metadata_type: str) -> list:
        query = self.sf.mdapi.ListMetadataQuery(type=metadata_type)
        query_response = self.sf.mdapi.list_metadata(query)
        return query_response
    
    def metadata_rename(self, metadata_type: str, previous_name: str, new_name: str): 
        """
        Renames the API name of SFDC metadata
        """
        getattr(self.sf.mdapi, metadata_type).rename(previous_name, new_name)
    

    ### ----------------------------------------------------------------------------------------------------
    ### Utility Functions
    ### ----------------------------------------------------------------------------------------------------
    
    def sf_api_call(self, action, parameters = {}, method = 'get', data = {}):
        """
        Helper function to make calls to Salesforce REST API.
        Parameters: action (the URL), URL params, method (get, post or patch), data for POST/PATCH.
        """
        headers = {
            'Content-type': 'application/json',
            'Accept-Encoding': 'gzip',
            'Authorization': 'Bearer %s' % self.session_id
        }
        if method == 'get':
            r = requests.request(method, 'https://'+self.instance+action, headers=headers, params=parameters, timeout=30)
        elif method in ['post', 'patch']:
            r = requests.request(method, 'https://'+self.instance+action, headers=headers, json=data, params=parameters, timeout=10)
        else:
            # other methods not implemented in this example
            raise ValueError('Method should be get or post or patch.')
#         print('Debug: API %s call: %s' % (method, r.url) )
        if r.status_code < 300:
            if method=='patch':
                return None
            else:
                return r.json()
        else:
            raise Exception('API error when calling %s : %s' % (r.url, r.content))


    def join(self, 
                dataframe, 
                rt_dataframe, 
                left_on="", 
                right_on="", 
                new_columns=[], 
                suffix = "_new"):
        """
        Joins and appends the Name and DeveloperName columns of record type to the dataframe.
        The dataframe must contain a column named "RecordTypeId" with the 18 digit SFID of the record type.
        """
        df_columns = dataframe.columns.to_list()
        if right_on not in new_columns:
            new_columns.insert(0,right_on)
        dataframe = pd.merge(dataframe, rt_dataframe[new_columns], how = 'inner', 
                                left_on = left_on,right_on = right_on, suffixes=('', suffix))
#         dataframe.drop([col for col in dataframe.columns if 'drop' in col], axis=1, inplace=True)
        return dataframe



from collections import OrderedDict
import pandas as pd


def recursive_walk(od_field: OrderedDict):
    """
    Recursively flattens each row the results of simple salesforce.
    Only works for bottom up queries.
    :param od_field: results returned by simple salesforce (multiple objects)
    :return: returns a flattened list of dictionaries
    """
    d = {}
    for k in od_field.keys():
        if isinstance(od_field[k], OrderedDict) & (k != 'attributes'):
            if 'attributes' in od_field[k].keys():
                ret_df = recursive_walk(od_field[k])
                d = {**d, **ret_df}
        else:
            if k != 'attributes':
                obj = od_field['attributes']['type'].replace(" ", ".")
                d[f'{obj}.{k}'] = od_field[k]
    return d

def transform_sf_result_set_rec(query_results: OrderedDict):
    """
    Recursively flattens the results of simple salesforce. It needs flattening when  selecting
    multiple objects.
    :param query_results:
    :return:
    """
    data = []
    for res in query_results:
        d = recursive_walk(res)
        data.append(d)
    data = pd.DataFrame(data)
    return data

def deep_merge_dictionaries(source, destination):
    """
    run me with nosetests --with-doctest file.py

    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> deep_merge_dictionaries(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
    True
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            deep_merge_dictionaries(value, node)
        else:
            destination[key] = value

    return destination


from sys import platform
import requests
import pandas as pd
import os
import json
from typing import List, Tuple, Optional, Union, Type, OrderedDict, Iterable
import numpy as np
import sys
import importlib.resources
from urllib.parse import urlparse
try:
    from custom_exceptions.veevanetwork_exceptions import *
except:
    from veevanetwork.custom_exceptions.veevanetwork_exceptions import *

class Vn:
    def __init__(self):
        self.os_platform: str = platform
        self.networkURL: str = None # https://verteonetwork.veevanetwork.com
        self.networkUserName: str = None
        self.networkPassword: str = None
        self.networkCountry: str = 'US'
        self.networkConnection: requests.models.Response = None
        self.networkLanguages: list = None # List of languages supported by this instance of Veeva Network
        self.networkObjects: dict = None # Dictionary of Network Objects and their properties
        self.networkObjectMetadata: dict = None # {'HCP': {'my_custom_field__c': {'fieldId': 'my_custom_field__c','type': {'dataType': 'STRING', 'discriminator': None}, 'labels': {'en': 'My Custom Field'}}}}
        self.networkReferenceTypes: dict = None # {'AddressAdminArea': {'type': 'AddressAdminArea','customerOwned': False,'inactive': False,'description': 'AddressAdminArea'}, 'AddressCBSA': {'type': 'AddressCBSA' ... }}
        self.networkReferenceValueMetadata: dict = None # Dictionary of Network Reference Value Metadata, including countries, reference codes, translated values, etc.
        self.sessionId: str = None
        self.networkId: str = None
        self.APIheaders: dict = None
        self.APIversionList: list = []
        self.LatestAPIversion: str = None
        self.network_references_all: pd.DataFrame = None
        self.network_DNS: str = None # verteonetwork.veevanetwork.com
        self.network_protocol: str = None # https

        #----------------------------------------------------------------
        # Load Network API Json
        self.network_api_json: dict = None
        try: # Try to load the network API json, to be used in a packaged distribution context
            with importlib.resources.open_text("veevanetwork", "network_api_v25.json") as file:
                self.network_api_json = json.load(file)
        except: # If the above fails, try to load the network API json, to be used in a development context
            with open('network_api_v25.json', encoding="utf-8") as file:
                self.network_api_json = json.load(file)
        self.network_api_parsed: dict = self.__parse_network_api_json(self.network_api_json)
        self.api_categories: list = list(self.network_api_parsed['item'].keys())

        self.api_change_request = self.network_api_parsed['item']['Change Request']
        self.api_custom_key = self.network_api_parsed['item']['Custom Key']
        self.api_entity = self.network_api_parsed['item']['Entity']
        self.api_hco = self.network_api_parsed['item']['HCO']
        self.api_hcp = self.network_api_parsed['item']['HCP']
        self.api_metadata = self.network_api_parsed['item']['Metadata']
        self.api_subscriptions = self.network_api_parsed['item']['Subscriptions']
        self.api_subscriptions_compliance = self.network_api_parsed['item']['Subscriptions - Compliance']
        self.api_subscriptions_source = self.network_api_parsed['item']['Subscriptions - Source']
        self.api_subscriptions_target = self.network_api_parsed['item']['Subscriptions - Target']
        self.api_events = self.network_api_parsed['item']['Events']
        self.api_match = self.network_api_parsed['item']['Match']
        self.api_suspect_match = self.network_api_parsed['item']['Suspect Match']
        self.api_authentication = self.network_api_parsed['item']['Authentication']
        self.api_retrieve_available_api_versions = self.network_api_parsed['item']['Retrieve Available API Versions']
        self.api_search = self.network_api_parsed['item']['Search']
        self.call = {
            "Change Request": {
                "change_request_create": self.change_request_create,
                "change_request_cancel": self.change_request_cancel,
                "change_request_retrieve": self.change_request_retrieve,
                "change_request_batch_retrieve": self.change_request_batch_retrieve,
                "change_request_update": self.change_request_update,
                "change_request_batch_update": self.change_request_batch_update,
                "change_request_process": self.change_request_process,
                "change_request_batch_process": self.change_request_batch_process,
                "change_request_search": self.change_request_search,
                "change_request_match": self.change_request_match,
                "change_request_batch_approve": self.change_request_batch_approve,
                "change_request_batch_reject": self.change_request_batch_reject
            },
            "Custom Key": {
                "custom_key_associate_to_entity": self.custom_key_associate_to_entity,
                "custom_key_associate_to_child": self.custom_key_associate_to_child,
                "custom_key_batch_associate_to_entities": self.custom_key_batch_associate_to_entities,
                "custom_key_batch_associate_to_children": self.custom_key_batch_associate_to_children,
                "custom_key_disassociate": self.custom_key_disassociate,
                "custom_key_batch_disassociate": self.custom_key_batch_disassociate
            },
            "Entity": {
                "entity_retrieve": self.entity_retrieve,
                "entity_retrieve_child": self.entity_retrieve_child,
                "entity_batch_retrieve": self.entity_batch_retrieve,
                "entity_batch_retrieve_children": self.entity_batch_retrieve_children
            },
            "HCO": {
                "hco_retrieve": self.hco_retrieve,
                "hco_associate_custom_key": self.hco_associate_custom_key,
                "hco_address_associate_custom_key": self.hco_address_associate_custom_key,
                "hco_license_associate_custom_key": self.hco_license_associate_custom_key,
                "hco_parenthco_associate_custom_key": self.hco_parenthco_associate_custom_key
            },
            "HCP": {
                "hcp_retrive": self.hcp_retrive,
                "hcp_associate_custom_key": self.hcp_associate_custom_key,
                "hcp_address_associate_custom_key": self.hcp_address_associate_custom_key,
                "hcp_license_associate_custom_key": self.hcp_license_associate_custom_key,
                "hcp_parenthco_associate_custom_key": self.hcp_parenthco_associate_custom_key
            },
            "Metadata": {
                "metadata_retrieve_available_api_versions": self.metadata_retrieve_available_api_versions,
                "metadata_retrieve_hashtags": self.metadata_retrieve_hashtags,
                "metadata_retrieve_object_types": self.metadata_retrieve_object_types,
                "metadata_retrieve_field": self.metadata_retrieve_field,
                "metadata_retrieve_field_details": self.metadata_retrieve_field_details,
                "metadata_retrieve_field_groups": self.metadata_retrieve_field_groups,
                "metadata_retrieve_reference_data_types": self.metadata_retrieve_reference_data_types,
                "metadata_retrieve_reference_data_type_details": self.metadata_retrieve_reference_data_type_details,
                "metadata_retrieve_reference_data_type_code_details": self.metadata_retrieve_reference_data_type_code_details
            },
            "Subscriptions": {
                "subscriptions_create_job": self.subscriptions_create_job,
                "subscriptions_retrieve_job": self.subscriptions_retrieve_job,
                "subscriptions_cancel_job": self.subscriptions_cancel_job,
                "subscriptions_retrieve_export_job_file": self.subscriptions_retrieve_export_job_file
            },
            "Subscriptions - Compliance": {
                "subscriptions_compliance_create_job": self.subscriptions_compliance_create_job,
                "subscriptions_compliance_retrieve_job": self.subscriptions_compliance_retrieve_job,
                "subscriptions_compliance_cancel_job": self.subscriptions_compliance_cancel_job
            },
            "Subscriptions - Source": {
                "subscriptions_source_create_job": self.subscriptions_source_create_job,
                "subscriptions_source_retrieve_job": self.subscriptions_source_retrieve_job,
                "subscriptions_source_cancel_job": self.subscriptions_source_cancel_job
            },
            "Subscriptions - Target": {
                "subscriptions_target_create_job": self.subscriptions_target_create_job,
                "subscriptions_target_retrieve_job": self.subscriptions_target_retrieve_job,
                "subscriptions_target_cancel_job": self.subscriptions_target_cancel_job
            },
            "Events": {
                "events_retrieve_merge": self.events_retrieve_merge,
                "events_retrieve_unmerge": self.events_retrieve_unmerge
            },
            "Match": {
                "match_retrieve": self.match_retrieve,
                "suspect_match_batch_reject_task": self.suspect_match_batch_reject_task,

            },
            "Suspect Match": {
                "suspect_match_batch_process": self.suspect_match_batch_process,
                "suspect_match_batch_create": self.suspect_match_batch_create,
                "suspect_match_batch_reject_task": self.suspect_match_batch_reject_task,
                "suspect_match_batch_retrieve": self.suspect_match_batch_retrieve
            },
            "Retrieve Available API Versions": {
                "retrieve_available_api_versions": self.retrieve_available_api_versions
            },
            "Search": {
                "search": self.search
            }
            }
    
    # =============================================================================
    # Authentication
    # =============================================================================

    # Refactored - MP 20220610
    def authenticate(
        self, 
        networkURL: str=None, 
        networkUserName: str=None, 
        networkPassword: str=None, 
        networkCountry: str=None, 
        networkId: str=None,
        sessionId: str=None,
        if_return: bool=False, 
        *args, **kwargs
    ) -> Optional[dict]:
        """
        Authenticates Veeva Network and retrieves the auth token.

        Example:
        authenticate using unpacked kwargs from the retrieve_credentials function
            authenticate_vn(**retrieve_credentials(platform, 'credentials.xlsx')[0])

        Return Example:
            {'sf': <simple_salesforce.api.Salesforce at 0x24a045c7b50>,
             'bulk': <salesforce_bulk.salesforce_bulk.SalesforceBulk at 0x24a045c7e20>,
             'sfMeta': <sfdclib.session.SfdcSession at 0x24a044b3400>,
             'tooling': <sfdclib.tooling.SfdcToolingApi at 0x24a045d17f0>,
             'session_id': '00D3F000000FZCq!AQYAQHYfSLYGI9cTyjDfxAAzYm.1uOKmNPXlKMW0sVz5ilIQ9ZwVTh6kOlaRuqfPuuzNnZNb3461sUGeUZ57ttE.GBawbt5h',
             'instance': 'cslbehring-core--devr01.my.salesforce.com',
             'sfMeta_is_connected': True,
             'bulk_api_sessionId': '00D3F000000FZCq!AQYAQHYfSLYGI9cTyjDfxAAzYm.1uOKmNPXlKMW0sVz5ilIQ9ZwVTh6kOlaRuqfPuuzNnZNb3461sUGeUZ57ttE.GBawbt5h'}

        Dependencies:
            import requests

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Authenticate
        
        """
        if (networkId and sessionId and networkURL):
            self.networkId = self.networkId if networkId is None else networkId
            self.sessionId = self.sessionId if sessionId is None else sessionId
        
        url_parse = urlparse(networkURL)
        if len(url_parse.scheme) == 0:
            self.network_protocol = 'https'
            if len(url_parse.path) > 0:
                self.network_DNS = url_parse.path
                self.networkURL = self.network_protocol + '://' + url_parse.path

        if len(url_parse.scheme) > 0:
            self.network_protocol = url_parse.scheme
            if len(url_parse.netloc) > 0:
                self.network_DNS = url_parse.netloc
                self.networkURL = url_parse.scheme + '://' + url_parse.netloc

        if (self.networkURL is None) or (len(self.networkURL) == 0):
            raise Exception('networkURL is required')


        # self.networkURL = self.networkURL if networkURL is None else networkURL
        self.networkUserName = self.networkUserName if networkUserName is None else networkUserName
        self.networkPassword = self.networkPassword if networkPassword is None else networkPassword
        self.networkCountry = self.networkCountry if networkCountry is None else networkCountry
        
        if (self.networkUserName and self.networkPassword and self.networkURL):
            self.networkConnection = requests.post(self.networkURL + '/api/v17.0/auth?username=' + self.networkUserName + '&password=' + self.networkPassword)
            if (self.networkConnection.json()['responseStatus'] == 'SUCCESS'):
                self.sessionId = self.networkConnection.json()['sessionId']
                self.userId = self.networkConnection.json()['userId']
                self.networkId = self.networkConnection.json()['networkId']
            else:
                raise Exception(self.networkConnection.json())

        
        self.APIheaders = {'authorization': self.sessionId}
        self.APIversionList = []
        
        # Error checking whether the required parameters are passed in
        # The check happens here because this is where all the self assignments has completed
        if (not (self.networkId and self.sessionId and self.networkURL)) and (not (self.networkUserName and self.networkPassword and self.networkCountry and self.networkURL)):
            raise Exception("Please provide either networkId, sessionId, and networkURL or networkUserName, networkPassword, and networkURL")
        
        for API in requests.get(self.networkURL +'/api', headers=self.APIheaders).json()['values'].keys():
            self.APIversionList.append(float(API.replace("v", "")))
        self.APIversionList.sort()
        self.LatestAPIversion = "v" + str(self.APIversionList[-1])
        

        if if_return:
            return {'networkURL':self.networkURL, 
                    'networkUserName':self.networkUserName, 
                    'networkPassword':self.networkPassword, 
                    'networkConnection':self.networkConnection, 
                    'sessionId':self.sessionId, 
                    'APIheaders':self.APIheaders, 
                    'APIversionList':self.APIversionList, 
                    'LatestAPIversion':self.LatestAPIversion}    

    # Refactored - MP 20220610
    def get_networkObjects(self) -> pd.DataFrame:
        """
        Returns a list of all objects in the Veeva Network.
        """
        self.authenticate()

        self.networkObjects = self.list_of_dicts_to_dict(requests.get(self.networkURL + '/api/' + self.LatestAPIversion + '/metadata/objectTypes',
                                      headers=self.APIheaders).json()['objectTypes'], 'name')
        return self.networkObjects

    # Refactored - MP 20220610
    def get_networkObjectMetadata(self) -> pd.DataFrame:
        self.authenticate()
        if self.networkObjects is None:
            self.get_networkObjects()

        processing_dict = {}
        __empty =[
            processing_dict.update({networkObject: 
                self.list_of_dicts_to_dict(requests.get(
                        self.networkURL + 
                        '/api/'+ self.LatestAPIversion +
                        '/metadata/fields?objectTypes=' + 
                        networkObject + '&details=full' + 
                        '&countries='+ self.networkCountry, 
                        headers=self.APIheaders).json()['attributes'],
                        'fieldId')}) 
                        for networkObject in self.networkObjects.keys()]
        self.networkObjectMetadata = processing_dict
        return self.networkObjectMetadata


    # Refactored - MP 20220612
    def process_networkReferenceCodeMetadata(
        self,
        referenceTypes: Optional[list] = None, # List of reference types to retrieve, leave empty (None) to retrieve all
        customerOwned: Optional[bool] = None, # True (customer owned only), False (ootb), None (customer owned and ootb)
        activeReferences: Optional[bool]= None, # True (active only), False (inactive only), None (active and inactive)
        languages: Optional[list]= None, # List of languages to retrieve, leave empty (None) to retrieve all
        countries: Optional[list]= None, # List of countries to retrieve, leave empty (None) to retrieve all
    ):
        """
        Processes Network Reference Metadata and returns a DataFrame of the reference codes based on the parameters.

        Parameters:
        referenceTypes: List of reference types to retrieve, leave empty (None) to retrieve all
        customerOwned (bool): Filter references by customerOwned status. Values: True (customerOwned only), False (ootb only), None (customerOwned and ootb)
        activeReferences (bool): Filter references by status. Values: True (active only), False (inactive only), None (active and inactive)
        languages (list): List of languages to return. Default: ['en']
        countries (list): List of countries to return. Default: ['US']

        Returns:
        DataFrame: A DataFrame of the reference codes based on the parameters.
        """

        if self.networkReferenceValueMetadata is None:
            self.get_networkReferenceValueMetadata()
        
        if languages is None:
            if self.networkLanguages is None:
                self.get_networkLanguages()
            languages = self.networkLanguages
            
        networkReferenceDF = pd.DataFrame(self.networkReferenceValueMetadata)
        networkReferenceCodeDataframe = networkReferenceDF.loc['reference_type_codes', :].dropna()
        reference_code_dict_unfiltered: dict[str, pd.DataFrame] = {}
        for index, values in zip(networkReferenceCodeDataframe.index, networkReferenceCodeDataframe.values):
            reference_code_dict_unfiltered.update({index: pd.DataFrame(self.list_of_dicts_to_dict(values, 'code'))})
        
        final_df = pd.DataFrame()
        for reference_type, reference_dataframe in zip(reference_code_dict_unfiltered.keys(), reference_code_dict_unfiltered.values()):
            if referenceTypes is None or reference_type in referenceTypes:
                activeReferences = None
                customerOwned = None
                activeReferences_filter = [True,False] if activeReferences is None else [not(activeReferences)]
                customerOwned_filter = [True,False] if customerOwned is None else [customerOwned]
                reference_dataframe = reference_dataframe.T.copy()
                filtered_customerOwned_and_active = reference_dataframe[
                                                (reference_dataframe['customerOwned'].isin(customerOwned_filter)) & 
                                                (reference_dataframe['inactive'].isin(activeReferences_filter))
                                                ]

                filtered_languages = pd.DataFrame()

                for language in languages:
                    filtered_languages = pd.concat([filtered_languages,
                                                pd.DataFrame(
                                                    filtered_customerOwned_and_active["values"]
                                                    .apply(lambda value: value[language] 
                                                    if value.keys()
                                                    .__contains__(language) 
                                                    else np.nan)
                                                    .dropna()
                                                    .rename(language))], axis=1)
                filtered_languages.insert(0, 'Network Code', filtered_languages.index)
                filtered_languages.insert(1, 'Reference Type', reference_type)
                filtered_languages.insert(2, 'Active Countries', pd.DataFrame(filtered_customerOwned_and_active["countries"]
                                                    .apply(lambda value: ";".join(sorted(value)) if (countries is None or len(set(value).intersection(set(countries)))>0 ) else np.nan).rename('countries')
                                                )['countries'])
                filtered_languages.insert(3, 'Definition', pd.DataFrame(
                                                    filtered_customerOwned_and_active["values"]
                                                    .apply(lambda value: value['en'] 
                                                    if value.keys()
                                                    .__contains__('en') 
                                                    else np.nan)
                                                    .dropna()
                                                    .rename('en'))['en'])
                filtered_languages.insert(4, 'Active?', filtered_customerOwned_and_active["inactive"] == False)
                filtered_languages.insert(5, 'Veeva Maintained?', filtered_customerOwned_and_active["customerOwned"] == False)             
                filtered_languages.reset_index(drop=True, inplace=True)
                filtered_languages.dropna(subset=['Active Countries'], inplace=True)
                filtered_languages.replace(np.nan, "", inplace=True)
                filtered_languages
                final_df = pd.concat([final_df, filtered_languages], axis=0)

        final_df.reset_index(drop=True, inplace=True)
        
        return final_df

    # To be deprecated, used by object_metadata_dataframe()
    def object_metadata(self):
        self.networkObjects = requests.get(self.networkURL + '/api/' + self.LatestAPIversion + '/metadata/objectTypes',
                                      headers=self.APIheaders).json()['objectTypes']
        networkObjectList = []
        for obj in self.networkObjects:
            if obj['status'] == 'ACTIVE':
                networkObjectList.append(obj['name'])
            else:
                pass
        object_metadata = pd.DataFrame([requests.get(self.networkURL + '/api/'+ self.LatestAPIversion +'/metadata/fields?objectTypes=' + networkObject + '&details=full' + '&countries='+ self.networkCountry, headers=self.APIheaders).json()['attributes'] 
                                              for networkObject in networkObjectList])
        object_metadata = object_metadata.transpose()
        object_metadata.columns = networkObjectList
        return object_metadata

    # To be deprecated, use get_networkObjectMetadata() instead
    def object_metadata_dataframe(self, object_metadata = None, attribute = None, *args):
        """
        Function that takes an input of data table with metadata values containing field for each row and objects for each column, 
        lists of input attributes available in the meta data (i.e. ['type','dataType']), 
        and optional arguments containing comma separated lists of additional attributes available in the meta data 
        
        Example:
            VeevaNetwork.object_metadata_dataframe(self.object_metadata(), ['fieldId'], ['type','dataType'], ['type','discriminator'], ['labels','en'],['maximumLength'])
            
        Return Example:
        (Dataframe)
            HCP.fieldId    HCP.type.dataType    HCP.type.discriminator    HCP.labels.en    HCP.maximumLength    HCO.fieldId      ...
            hcp_type__v    REFERENCE            HCPType                   HCP Type         100.0                340B_eligible__v ...
        
        """
        self.authenticate()
        
        object_metadata = self.object_metadata() if object_metadata is None else object_metadata
        
        object_metadata_dataframe = pd.DataFrame()
        attribute1 = attribute[0]
        attribute2 = attribute[1] if len(attribute) > 1 else ""
        for networkObjectName, networkObjectMetaData in object_metadata.iteritems():
            attributeList = []
            for fieldMetaData in networkObjectMetaData:
                try:
                    if attribute2 == "":
                        attributeList.append(fieldMetaData[attribute1])
                    else:
                        attributeList.append(fieldMetaData[attribute1][attribute2])
                except TypeError:
                    continue
    #        object_metadata_dataframe[networkObjectName + '.' + attribute1 + (("." + attribute2) if attribute2 != "" else "")] = pd.Series(attributeList, name = networkObjectName)
            object_metadata_dataframe = pd.concat([object_metadata_dataframe,pd.Series(attributeList, name = networkObjectName + '.' + attribute1 + (("." + attribute2) if attribute2 != "" else "")).to_frame()], ignore_index=False, axis=1)


            # parse arguments
            if args:
                for arg in args:
                    argAttribute1 = arg[0]
                    argAttribute2 = arg[1] if len(arg)>1 else ""
                    argAttributeList = []
                    for fieldMetaData in networkObjectMetaData:
                        try:
                            if argAttribute2 == "":
                                argAttributeList.append(fieldMetaData[argAttribute1])
                            else:
                                argAttributeList.append(fieldMetaData[argAttribute1][argAttribute2])
                        except:
                            argAttributeList.append("")
                            continue
    #                object_metadata_dataframe[networkObjectName + '.' + argAttribute1 + (("." + argAttribute2) if argAttribute2 != "" else "")] = pd.Series(argAttributeList, name = networkObjectName)
                    object_metadata_dataframe = pd.concat([object_metadata_dataframe,pd.Series(argAttributeList, name = networkObjectName + '.' + argAttribute1 + (("." + argAttribute2) if argAttribute2 != "" else "")).to_frame()], ignore_index=False, axis=1)
            else:
                continue
        return object_metadata_dataframe

    # Refactored - MP 20220611
    def get_networkReferenceValueMetadata(self):
        """
        Retruns a Dictionary of Network Reference Value Metadata, including countries, reference codes, translated values, etc.
        """
        self.authenticate()
        self.networkReferenceValueMetadata = self.list_of_dicts_to_dict(requests.get(self.networkURL + '/api/' + self.LatestAPIversion + '/metadata/reference_values?includeCodes=True', headers = self.APIheaders).json()['reference_type_values'],'type')
        return self.networkReferenceValueMetadata

    # Deprecated
    def reference_value_dataframe(self):
        """
        Retrieve all reference values in Network  
        ** formula can be enhanced or take input parameters so it doesn't query every single reference alias

        """
        self.authenticate()

        self.networkReferenceTypes = self.list_of_dicts_to_dict(requests.get(self.networkURL + '/api/' + self.LatestAPIversion + '/metadata/reference_values?includeCodes=True', headers = self.APIheaders).json()['reference_type_values'],'type')
        reference_value_dataframe = pd.DataFrame()
        for x in self.networkReferenceTypes.keys():
            try:
                referenceCall = requests.get(self.networkURL + '/api/'+ self.LatestAPIversion + '/metadata/reference_values/' + x + '?countries='+ self.networkCountry, headers = self.APIheaders).json()['reference_type_codes']
                newColumn = pd.Series([x['values']['en'] for x in referenceCall],name = str(x) + " Value")
                
                newColumn2 = pd.Series([x['code'] for x in referenceCall], name = x)
                reference_value_dataframe.insert(len(reference_value_dataframe.columns), column = x, value = newColumn2)
                reference_value_dataframe.insert(len(reference_value_dataframe.columns), column = str(x) + " Value", value = newColumn)
            except:
                pass
        return reference_value_dataframe

    def extract_network_reference_table(self):
        self.authenticate()
        
        references_all = pd.DataFrame(requests.get(self.networkURL + '/api/'+ self.LatestAPIversion + '/metadata/reference_values?includeCodes=true&countries='+ self.networkCountry, headers = self.APIheaders).json()['reference_type_values'])
        network_reference_table = pd.DataFrame()
        self.network_references_all = references_all
        for row in references_all['reference_type_codes']:
            if isinstance(row, list):
                network_reference_table = pd.concat([network_reference_table, pd.json_normalize(row)], axis=1)
            else:
                continue
        network_reference_table.fillna('',inplace=True)
        return network_reference_table
    
    # =============================================================================
    # WIP Functions
    # =============================================================================


    # =============================================================================
    # Change Request
    # =============================================================================

    def change_request_create(self, *args, **kwargs):
        """
        This API enables you to create a Network change request to add or update an HCP or HCO and related entities (including addresses, licenses, or parentHCOs) for **gray** data only.

        Change requests created using the API against orange data will be rejected.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Createachangerequest

        """
        
        return self.api_call(self.api_change_request['item']['Create Change Request'], *args, **kwargs)
        
    def change_request_cancel(self, *args, **kwargs):
        """
        This API enables you to cancel a change request by providing the change request ID.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Cancelchangerequest

        """
        return self.api_call(self.api_change_request['item']['Cancel Change Request'], *args, **kwargs)

    def change_request_retrieve(self, *args, **kwargs):
        """
        This API enables you to retrieve response information for the create, update, and merge change requests submitted by a client application.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveChangeRequest

        """
        return self.api_call(self.api_change_request['item']['Retrieve Change Request'], *args, **kwargs)
        
    def change_request_batch_retrieve(self, *args, **kwargs):
        """
        This API enables you to obtain information about multiple change requests through the API.

        Documenatation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Batchretrievechangerequest  

        """
        return self.api_call(self.api_change_request['item']['Change Request Batch Retrieve'], *args, **kwargs)
        
    def change_request_update(self, *args, **kwargs):
        """
        This API enables you to update an unprocessed change request through the API.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#UpdateChangeRequest

        """
        return self.api_call(self.api_change_request['item']['Change Request Update'], *args, **kwargs)

    def change_request_batch_update(self, *args, **kwargs):
        """
        This API enables you to update multiple unprocessed change requests.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchUpdateChangeRequests

        """
        return self.api_call(self.api_change_request['item']['Change Request Batch Update'], *args, **kwargs)

    def change_request_process(self, *args, **kwargs):
        """
        This API enables you to process an unprocessed change request through the API.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#ProcessChangeRequest

        """
        return self.api_call(self.api_change_request['item']['Change Request Process'], *args, **kwargs)

    def change_request_batch_process(self, *args, **kwargs):
        """
        This API enables you to process multiple unprocessed change requests.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchProcessChangeRequests
        """
        return self.api_call(self.api_change_request['item']['Change Request Batch Process'], *args, **kwargs)

    def change_request_search(self, *args, **kwargs):
        """
        This API enables you to retrieve all change requests that match a specified search criteria.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#ChangeRequestSearch

        """
        return self.api_call(self.api_change_request['item']['Change Request Search'], *args, **kwargs)

    def change_request_match(self, *args, **kwargs):
        """
        This API enables you to match a request to an existing entity. You can use either the Veeva ID or custom key of an existing entity to match the request against.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#ChangeRequestMatch

        """
        return self.api_call(self.api_change_request['item']['Change Request Match'], *args, **kwargs)

    def change_request_batch_approve(self, *args, **kwargs):
        """
        This API enables you to bulk approve up to 500 change requests.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchApproveChangeRequests

        """
        return self.api_call(self.api_change_request['item']['Change Request Batch Approve'], *args, **kwargs)

    def change_request_batch_reject(self, *args, **kwargs):
        """
        This API enables you to bulk reject up to 500 change requests.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchRejectChangeRequests

        """
        return self.api_call(self.api_change_request['item']['Change Request Batch Reject'], *args, **kwargs)
    
    # =============================================================================
    # Custom Key
    # =============================================================================

    def custom_key_associate_to_entity (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCPs or HCOs are downloaded from Network without going through the full change request process.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytoentity

        """
        return self.api_call(self.api_custom_key['item']['Associate Custom Key to Entity'], *args, **kwargs)

    def custom_key_associate_to_child (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new children (address, license, or parent HCOs) are downloaded from Network without going through the full change request process.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytochildren

        """
        return self.api_call(self.api_custom_key['item']['Associate Custom Key to Children'], *args, **kwargs)

    def custom_key_batch_associate_to_entities (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers whenever new HCPs or HCOs are downloaded from Network without going through the full change request process.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Batchassociatecustomkeystoentities

        """
        return self.api_call(self.api_custom_key['item']['Batch Associate Custom Keys to Entities'], *args, **kwargs)

    def custom_key_batch_associate_to_children (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new children (address, license, or parent HCOs) are downloaded from Network without going through the full change request process.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Batchassociatecustomkeytochildren

        """
        return self.api_call(self.api_custom_key['item']['Batch Associate Custom Keys to Children'], *args, **kwargs)

    def custom_key_disassociate (self, *args, **kwargs):
        """
        This API enables you to deactivate external key identifiers for any entity (an HCP, HCO, Address, License, or ParentHCO) in Network without going through the full change request process.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Disassociatecustomkey

        """
        return self.api_call(self.api_custom_key['item']['Disassociate Custom Key'], *args, **kwargs)

    def custom_key_batch_disassociate (self, *args, **kwargs):
        """
        This API enables you to inactivate external key identifiers for any entity (HCP, HCO, Address, License, or ParentHCO) in Network without going through the full change request process.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Batchdisassociatecustomkey

        """
        return self.api_call(self.api_custom_key['item']['Batch Disassociate Custom Key'], *args, **kwargs)


    # =============================================================================
    # Entity
    # =============================================================================

    def entity_retrieve (self, *args, **kwargs):
        """
        This API enables you to obtain information on any entity without identifying the specific entity type. It is only used to retrieve information from Network using the GET method.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrieveentity

        """
        return self.api_call(self.api_entity['item']['Retrieve Entity'], *args, **kwargs)

    def entity_retrieve_child (self, *args, **kwargs):
        """
        This API enables you to retrieve child entity information; for example address or license details, for the Network ID provided without identifying the specific entity type.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievechildentity

        """
        return self.api_call(self.api_entity['item']['Retrieve Child Entity'], *args, **kwargs)

    def entity_batch_retrieve (self, *args, **kwargs):
        """
        This API enables you to retrieve entity details directly from Network. It is only used to retrieve information using the GET method. To update or delete entity data, you must use the change request APIs.

        The entities you can retrieve include HCPs and HCOs, and details are returned for all corresponding child entities: addresses, licenses, parent HCOs, and custom keys.
        
        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Batchretrieveentities

        """
        return self.api_call(self.api_entity['item']['Batch Retrieve Entities'], *args, **kwargs)

    def entity_batch_retrieve_children (self, *args, **kwargs):
        """
        This API enables you to obtain information on child entities without identifying the specific entity type. Users are only allowed to retrieve (GET) information from Network.

        All other operations (POST and DELETE) are restricted and can only be performed by submitting a change request using the change request APIs.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Batchretrievechildentities

        """
        return self.api_call(self.api_entity['item']['Batch Retrieve Child Entities'], *args, **kwargs)


    # =============================================================================
    # HCO
    # =============================================================================

    def hco_retrieve (self, *args, **kwargs):
        """
        This API enables you to retrieve information about an HCO. Information you can retrieve for an HCO includes the HCO, address, license, and parent HCO information (including their custom keys) for the HCO vid_key you provide.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveHCO

        """
        return self.api_call(self.api_hco['item']['Retrieve HCO'], *args, **kwargs)

    def hco_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCOs are downloaded from Network without going through the full change request process. This API associates the external identifier you submit to the HCO vid_key you provide.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#AssociatecustomkeytoHCO

        """
        return self.api_call(self.api_hco['item']['Associate Custom Key to HCO'], *args, **kwargs)

    def hco_address_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCO child objects (addresses) 
        are downloaded from Network without going through the full change request process. 
        This API associates the external identifier you submit to the child object key you provide.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytoachildobject

        """
        return self.api_call(self.api_hco['item']['Associate Custom Key to HCO Address'], *args, **kwargs)

    def hco_license_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCO child objects (licenses) 
        are downloaded from Network without going through the full change request process. 
        This API associates the external identifier you submit to the child object key you provide.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytoachildobject

        """
        return self.api_call(self.api_hco['item']['Associate Custom Key to HCO License'], *args, **kwargs)

    def hco_parenthco_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCO child objects (parent HCOs) 
        are downloaded from Network without going through the full change request process. 
        This API associates the external identifier you submit to the child object key you provide.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytoachildobject

        """
        return self.api_call(self.api_hco['item']['Associate Custom Key to HCO ParentHCO'], *args, **kwargs)

    # =============================================================================
    # HCP
    # =============================================================================
    def hcp_retrive (self, *args, **kwargs):
        """
        This API enables you to retrieve information about an HCP including the HCP itself, address, license, and parent HCO information (including their custom keys) for the HCP vid_key you provide.

        This API downloads the record for the specified entity from Veeva OpenData to your customer instance.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveHCP

        """
        return self.api_call(self.api_hcp['item']['Retrieve HCP'], *args, **kwargs)

    def hcp_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCPs are downloaded from Network without going through the full change request process. This API associates the external identifier you submit to the HCP vid_key you provide.

        This API requires system administrator or API-only permissions.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#AssociatecustomkeytoHCP

        """
        return self.api_call(self.api_hcp['item']['Associate Custom Key to HCP'], *args, **kwargs)

    def hcp_address_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCP child objects (addresses) are downloaded from Network without going through the full change request process. This API associates the external identifier you submit to the child object key you provide.

        This API requires system administrator or API-only permissions.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytoachildobject1

        """
        return self.api_call(self.api_hcp['item']['Associate Custom Key to HCP Address'], *args, **kwargs)

    def hcp_license_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCP child objects (licenses) are downloaded from Network without going through the full change request process. This API associates the external identifier you submit to the child object key you provide.

        This API requires system administrator or API-only permissions.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytoachildobject1
        
        """
        return self.api_call(self.api_hcp['item']['Associate Custom Key to HCP License'], *args, **kwargs)

    def hcp_parenthco_associate_custom_key (self, *args, **kwargs):
        """
        This API enables you to submit external key identifiers when new HCP child objects (parent HCOs) are downloaded from Network without going through the full change request process. This API associates the external identifier you submit to the child object key you provide.

        This API requires system administrator or API-only permissions.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Associatecustomkeytoachildobject1
        
        """
        return self.api_call(self.api_hcp['item']['Associate Custom Key to HCP ParentHCO'], *args, **kwargs)

    # =============================================================================
    # Metadata
    # =============================================================================
    def metadata_retrieve_available_api_versions (self, *args, **kwargs):
        """
        This enables you to retrieve summary information about each API version available in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveavailableAPIversions
        
        """
        return self.api_call(self.api_metadata['item']['Retrieve Available API Versions'], *args, **kwargs)
        
    def metadata_retrieve_hashtags (self, *args, **kwargs):
        """
        This API enables you to retrieve the list of hashtags available in a Network instance.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievehashtags

        """
        return self.api_call(self.api_metadata['item']['Retrieve hashtags'], *args, **kwargs)

    # TODO: rename/refactor method.
    def get_networkLanguages(self) -> list:
        """
        This API enables you to retrieve the list of reference data languages available in a Network instance.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievelanguages

        """

        self.authenticate()
        self.networkLanguages = list(self.list_of_dicts_to_dict(
            requests.get(
                self.networkURL + 
                '/api/' + 
                self.LatestAPIversion + 
                '/metadata/languages', 
                headers=self.APIheaders).json()['languages'], 
                'name'
                ).keys())
        return self.networkLanguages

    def metadata_retrieve_object_types (self, *args, **kwargs):
        """
        This API enables you to retrieve the list of object types available in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrieveobjecttypesmetadata

        """
        return self.api_call(self.api_metadata['item']['Retrieve Object Types Metadata'], *args, **kwargs)

    def metadata_retrieve_field (self, *args, **kwargs):
        """
        This API enables you to retrieve detailed or summary information about the fields on each entity in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievefieldsmetadata

        """
        return self.api_call(self.api_metadata['item']['Retrieve Fields Metadata'], *args, **kwargs)

    def metadata_retrieve_field_details (self, *args, **kwargs):
        """
        This API enables you to retrieve detailed information about the fields on each entity in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievefielddetailsmetadata

        """
        return self.api_call(self.api_metadata['item']['Retrieve Field Details Metadata'], *args, **kwargs)

    def metadata_retrieve_field_groups (self, *args, **kwargs):
        """
        This API enables you to retrieve detailed information about the field groups available in Network. These field groups are used by the CRM bridge when retrieving and displaying information from Network in CRM.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievefieldgroupsmetadata

        """
        return self.api_call(self.api_metadata['item']['Retrieve Field Groups Metadata'], *args, **kwargs)

    def metadata_retrieve_reference_data_types (self, *args, **kwargs):
        """
        This API enables you to retrieve information about reference data types in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievereferencedatatypesmetadata

        """
        return self.api_call(self.api_metadata['item']['Retrieve Reference Data Types Metadata'], *args, **kwargs)

    def metadata_retrieve_reference_data_type_details (self, *args, **kwargs):
        """
        This API enables you to retrieve detailed information about reference data types in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievereferencedatatypedetailsmetadata

        """
        return self.api_call(self.api_metadata['item']['Retrieve Reference Data Type Details Metadata'], *args, **kwargs)

    def metadata_retrieve_reference_data_type_code_details (self, *args, **kwargs):
        """
        This API enables you to retrieve detailed information about reference data type codes in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievereferencedatatypecodedetailsmetadata

        """
        return self.api_call(self.api_metadata['item']['Retrieve Reference Data Type Code Details Metadata'], *args, **kwargs)

    # =============================================================================
    # Subscriptions
    # =============================================================================
    def subscriptions_create_job (self, *args, **kwargs):
        """
        This API enables you to create a subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Createasubscriptionjob

        """
        return self.api_call(self.api_subscriptions['item']['Create Subscription Job'], *args, **kwargs)

    def subscriptions_retrieve_job (self, *args, **kwargs):
        """
        This API enables you to retrieve the status of a source or target subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrieveasubscriptionjobstatus

        """
        return self.api_call(self.api_subscriptions['item']['Retrieve Subscription Job'], *args, **kwargs)

    def subscriptions_cancel_job (self, *args, **kwargs):
        """
        This API enables you to cancel a subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Cancelasubscriptionjob

        """
        return self.api_call(self.api_subscriptions['item']['Cancel Subscription Job'], *args, **kwargs)

    def subscriptions_retrieve_export_job_file (self, *args, **kwargs):
        """
        This API enables you to retrieve the artifacts (file contents) of a target subscription job.
        AKA: Retrieve target subscription artifact

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievetargetsubscriptionartifact
        
        """
        return self.api_call(self.api_subscriptions['item']['Retrieve Export Job File'], *args, **kwargs)

    # =============================================================================
    # Subscriptions - Compliance
    # =============================================================================
    def subscriptions_compliance_create_job (self, *args, **kwargs):
        """
        This API enables you to create a compliance subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Createacompliancesubscriptionjob

        """
        return self.api_call(self.api_subscriptions_compliance['item']['Create Compiance Subscription Job'], *args, **kwargs)

    def subscriptions_compliance_retrieve_job (self, *args, **kwargs):
        """
        This API enables you to retrieve the status of a source or target subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrieveacompliancesubscriptionjobstatus
        
        """
        return self.api_call(self.api_subscriptions_compliance['item']['Retrieve Compliance Subscription Job'], *args, **kwargs)

    def subscriptions_compliance_cancel_job (self, *args, **kwargs):
        """
        This API enables you to cancel a compliance subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Cancelacompliancesubscriptionjob

        """
        return self.api_call(self.api_subscriptions_compliance['item']['Cancel Compliance Subscription Job'], *args, **kwargs)

    # =============================================================================
    # Subscriptions - Source
    # =============================================================================
    def subscriptions_source_create_job (self, *args, **kwargs):
        """
        This API enables you to create a source subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Createasourcesubscriptionjob

        """
        return self.api_call(self.api_subscriptions_source['item']['Create Source Subscription Job'], *args, **kwargs)

    def subscriptions_source_retrieve_job (self, *args, **kwargs):
        """
        This API enables you to retrieve the status of a source subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrieveasourcesubscriptionjobstatus

        """
        return self.api_call(self.api_subscriptions_source['item']['Retrieve Source Subscription Job'], *args, **kwargs)

    def subscriptions_source_cancel_job (self, *args, **kwargs):
        """
        This API enables you to cancel a source subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Cancelasourcesubscriptionjob
        
        """
        return self.api_call(self.api_subscriptions_source['item']['Cancel Source Subscription Job'], *args, **kwargs)

    # =============================================================================
    # Subscriptions - Target
    # =============================================================================
    def subscriptions_target_create_job (self, *args, **kwargs):
        """
        This API enables you to create a target subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Createatargetsubscriptionjob

        """
        return self.api_call(self.api_subscriptions_target['item']['Create Target Subscription Job'], *args, **kwargs)

    def subscriptions_target_retrieve_job (self, *args, **kwargs):
        """
        This API enables you to retrieve the status of a source or target subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrieveatargetsubscriptionjobstatus

        """
        return self.api_call(self.api_subscriptions_target['item']['Retrieve Target Subscription Job'], *args, **kwargs)

    def subscriptions_target_cancel_job (self, *args, **kwargs):
        """
        This API enables you to cancel a target subscription job.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Cancelatargetsubscriptionjob
        """
        return self.api_call(self.api_subscriptions_target['item']['Cancel Target Subscription Job'], *args, **kwargs)

    # =============================================================================
    # Events
    # =============================================================================
    def events_retrieve_merge (self, *args, **kwargs):
        """
        This API enables you to retrieve the results of merge events that occurred in your Network instance.

        Merges initiated by Veeva OpenData on a master instance are included if the surviving and losing records of the merge have been downloaded to your instance.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveMerges

        """
        return self.api_call(self.api_events['item']['Retrieve Merge Events'], *args, **kwargs)

    def events_retrieve_unmerge (self, *args, **kwargs):
        """
        This API enables you to retrieve the results of unmerge events that occurred in your Network instance.

        Unmerges include events that occurred in your Network instance; only customer (gray) records are reported.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveUnmerges

        """
        return self.api_call(self.api_events['item']['Retrieve Unmerge Events'], *args, **kwargs)

    # =============================================================================
    # Match
    # =============================================================================
    def match_retrieve (self, *args, **kwargs):
        """
        This API enables you to match data immediately for a single record using match rules from the specified Network instance.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveMatches

        """
        return self.api_call(self.api_match['item']['Retrieve Matches'], *args, **kwargs)

    # =============================================================================
    # Suspect Match
    # =============================================================================
    def suspect_match_batch_process (self, *args, **kwargs):
        """
        This API enables enables you to process multiple unprocessed suspect matches.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchProcessSuspectMatch

        """
        return self.api_call(self.api_suspect_match['item']['Batch Process Suspect Match'], *args, **kwargs)

    def suspect_match_batch_create (self, *args, **kwargs):
        """
        This API enables enables you to create multiple suspect matches.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchCreateSuspectMatch

        """
        return self.api_call(self.api_suspect_match['item']['Batch Create Suspect Match'], *args, **kwargs)

    def suspect_match_batch_reject_task (self, *args, **kwargs):
        """
        This API enables enables you to reject multiple suspect match tasks.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchRejectSuspectMatch

        """
        return self.api_call(self.api_suspect_match['item']['Batch Reject Suspect Match Task'], *args, **kwargs)

    def suspect_match_batch_retrieve (self, *args, **kwargs):
        """
        This API enables enables you to retrieve information about multiple suspect matches.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#BatchRetrieveSuspectMatch

        """
        self.api_call(suspect_match['item']['Batch Retrieve Suspect Match'], *args, **kwargs)

    # =============================================================================
    # Retrieve Available API Versions
    # =============================================================================
    def retrieve_available_api_versions (self, *args, **kwargs):
        """
        This enables you to retrieve summary information about each API version available in Network.

        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#RetrieveavailableAPIversions

        """
        return self.api_call(self.api_retrieve_available_api_versions, *args, **kwargs)
    # =============================================================================
    # Search
    # =============================================================================
    def search (self, *args, **kwargs):
        """
        Search calls enable you to construct simple, yet powerful searches to retrieve data from Network.

        Calls through the Search API pass a query string in an expression that specifies the search text and specific parameters to get the intended set of entities from Network. Search results are ranked according to closeness to the search terms specified.
        
        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#search
        
        """
        return self.api_call(self.api_search, *args, **kwargs)
    # =============================================================================
    # System Settings
    # =============================================================================
    def system_settings_retrieve (self):
        """
        This API enables you to retrieve the value for the geolocation system setting.

        Currently only api.search.geolocation.countries is supported.
        
        Documentation:
        https://developer.veevanetwork.com/API_reference/API_reference.htm#Retrievesystemsettings
        
        """
        result = requests.get(
            url="https://demo.vdmdev.com/api/v26.0/systemSettings/api.search.geolocation.countries", 
            headers=self.APIheaders).json()
        return result
        

    # =============================================================================
    # Utilities
    # =============================================================================

    def api_call (
        self,
        api: dict, 
        body: dict = None, 
        variables: dict = None, 
        queries: dict = None, 
        if_print: bool=False,  # If True, prints each component of the api call, the api call will still execute unless if_debug or if_example paramters are set to True.
        if_debug: bool=False, # If True, the api call does not execute, but instead returns the api call as a string for debugging
        if_example: bool=False # If True, the api call does not execute and returns examples of any body, variables, or queries.
    ):

        examples = {}

        result = {
            'api_body' : body,
            'api_variables' : variables,
            'api_queries' : queries
        }

        request_params = {}

        def print_if(text):
            if if_print:
                print(text)

        if api.keys().__contains__('name'):
            print_if('\n- Name: ')
            print_if(api['name'])
            result['api_name'] = api['name'] ### Updating Result Name

        if api.keys().__contains__('protocolProfileBehavior'):
            print_if('\n- protocolProfileBehavior: ')
            if api['protocolProfileBehavior'].keys().__contains__('disableBodyPruning'):
                print_if('- disableBodyPruning: ')
                print_if(api['protocolProfileBehavior']['disableBodyPruning'])
                result['api_protocolProfileBehavior_disableBodyPruning'] = api['protocolProfileBehavior']['disableBodyPruning'] ### Updating Result disableBodyPruning

        if api.keys().__contains__('request'):

            ### =============================================================================
            ### Parsing Header
            ### =============================================================================

            if api['request'].keys().__contains__('header'):
                if isinstance(api['request']['header'], dict):
                    print_if("\n- Request header keys: ")
                    print_if(api['request']['header'].keys())
                    result['api_header'] = {}
                    request_params['headers'] = {}
                    for header_key, header_value in api['request']['header'].items():
                        print_if("\n- Request header key:")
                        print_if(header_key)
                        if isinstance(header_value, dict):
                            if header_value.keys().__contains__('value'):
                                result['api_header'].update({header_key: self.sub_request_params(api['request']['header'][header_key]['value'])}) ### Updating Result API Header
                                request_params['headers'].update({header_key: self.sub_request_params(api['request']['header'][header_key]['value'])}) ### Updating Request Header

                            for key, value in header_value.items():
                                print_if("\n- Request header value dict key and value:")
                                print_if(key)
                                print_if(value)
                        else:
                            print_if("Request Header Value: ")
                            print_if(header_value)
                            raise NetworkAPIHeaderValueNotFound

            ### =============================================================================
            ### Parsing Method
            ### =============================================================================

            if api['request'].keys().__contains__('method'):
                print_if("\n- Request method: ")
                print_if(api['request']['method'])

                match api['request']['method']:
                    case 'POST':
                        result['api_method'] = requests.post # Updating Result Method
                        print_if("\n- Request Method Post:")
                        print_if("requests.post")
                    case 'GET':
                        result['api_method'] = requests.get # Updating Result Method
                        print_if("\n- Request Method Get:")
                        print_if("requests.get")
                    case 'PUT':
                        result['api_method'] = requests.put # Updating Result Method
                        print_if("\n- Request Method Put:")
                        print_if("requests.put")
                    case 'DELETE':
                        result['api_method'] = requests.delete # Updating Result Method
                        print_if("\n- Request Method Delete:")
                        print_if("requests.delete")
                    case _:
                        print_if("\n- Request Method Unknown:")
                        print_if("unknown method")
                        raise NetworkAPIRequestMethodNotFound

            ### =============================================================================
            ### Parsing Body
            ### =============================================================================

            if (api['request'].keys().__contains__('body') == True) and \
                (api['request']['body'].keys().__contains__('raw') == True) and \
                (api['request']['body']['raw'] != ''):
                print_if("\n- Request body raw: ")
                print_if(json.loads(api['request']['body']['raw']))
                examples['body'] = json.loads(api['request']['body']['raw']) ### Updating Result API Body
                # If the body parameter is not provided in an API call that requires a body. Raise Error.
                if body is None and if_example==False:
                    raise NetworkAPIRequestBodyNotFound
                else:
                    request_params['data']=json.dumps(body)
            else:
                print_if("\n- Request body raw: ")
                print_if("No Body Found")
            
            ### =============================================================================
            ### Parsing URL
            ### =============================================================================

            if api['request'].keys().__contains__('url'):
                print_if("\n- Request url keys: ")
                print_if(api['request']['url'].keys())

                ### =============================================================================
                ### Parsing Raw URL
                ### =============================================================================

                if api['request']['url'].keys().__contains__('raw'):
                    print_if("\n- Request url raw: ")
                    print_if(api['request']['url']['raw'])
                    result['api_url'] = self.sub_request_params(api['request']['url']['protocol'] + "://" + api['request']['url']['host'][0] + "/" + "/".join(api['request']['url']['path'])) ### Updating Result API URL
                    request_params['url'] = self.sub_request_params(api['request']['url']['protocol'] + "://" + api['request']['url']['host'][0] + "/" + "/".join(api['request']['url']['path'])) ### Updating Request URL
                else:
                    raise NetworkAPIRequestURLNotFound

                
                ### =============================================================================
                ### Parsing URL Protocol
                ### =============================================================================
                
                if api['request']['url'].keys().__contains__('protocol'):
                    print_if("\n- Request url protocol: ")
                    # always "https"
                    print_if(api['request']['url']['protocol'])

                ### =============================================================================
                ### Parsing URL Host
                ### =============================================================================
                
                if api['request']['url'].keys().__contains__('host'):
                    print_if("\n- Request url host: ")
                    # always ['{{DNS}}]
                    print_if(api['request']['url']['host'][0])

                ### =============================================================================
                ### Parsing URL Path
                ### =============================================================================

                if api['request']['url'].keys().__contains__('path'):
                    print_if("\n- Request url path: ")
                    # A list of path components. i.e. ['api', '{{version}}', 'hcos', ':vid_key', 'addresses', ':address_key', 'custom_keys']
                    print_if(api['request']['url']['path'])

                ### =============================================================================
                ### Parsing URL Variable
                ### =============================================================================

                if api['request']['url'].keys().__contains__('variable'):
                    if variables is None and if_example==False:
                        raise NetworkAPIRequestVariableNotFound
                    elif if_example==False:
                        for variable in variables.keys():
                            result['api_url'] = result['api_url'].replace(":" + variable, variables[variable]['value'])
                            request_params['url'] = request_params['url'].replace(":" + variable, variables[variable]['value'])

                    if isinstance(api['request']['url']['variable'], dict):
                        print_if("\n- Request url variable keys: ")
                        print_if(api['request']['url']['variable'].keys())
                        variables: dict = api['request']['url']['variable']
                        examples['variables'] = variables ### Updating Result API URL Variable Examples
                        
                        for key, value in variables.items():
                            if isinstance(value, dict):
                                print_if("\n- Request url variable key and value dict: ")
                                print_if(key)
                                print_if(value)
                            else:
                                print_if("\n- Request url variable value: ") 
                                print_if(str(value))

                ### =============================================================================
                ### Parsing URL Query
                ### =============================================================================

                if api['request']['url'].keys().__contains__('query'):
                    if queries is None and if_example==False:
                        raise NetworkAPIRequestQueryNotFound
                    elif if_example==False:
                        result['api_url'] = result['api_url'] + '?'
                        request_params['url'] = request_params['url'] + '?'
                        for key, value in queries.items():
                            if result['api_url'][-1] == '?':
                                result['api_url'] = result['api_url'] + key + '=' + value['value']
                                request_params['url'] = request_params['url'] + key + '=' + value['value']
                            else:
                                result['api_url'] = result['api_url'] + '&' + key + '=' + value['value']
                                request_params['url'] = request_params['url'] + '&' + key + '=' + value['value']

                    print_if("\n- Request url query keys: ")
                    print_if(api['request']['url']['query'].keys())
                    queries: dict = api['request']['url']['query'] 
                    examples['query'] = queries ### Updating Result API URL Query Examples
                    for key, value in queries.items():
                        print_if("\n- Request url query key and value: ")
                        if isinstance(value, dict):
                            print_if(key)
                            print_if(value)
                        else:
                            print_if(str(value))

            ### =============================================================================
            ### Parsing Description
            ### =============================================================================

            if api['request'].keys().__contains__('description'):
                print_if("\n- Request Description: ")
                print_if(api['request']['description'])
                examples['description'] = api['request']['description'] ### Updating Description

        if api.keys().__contains__('response'):
            print_if(api['response'])

        if (not if_debug) & (not if_example) & (result.keys().__contains__('api_method')):
            response = result['api_method'](**request_params)

        if if_debug & if_example:
            return result, examples, request_params
        if if_debug:
            return result, request_params
        if if_example:
            return examples
        return response


    @staticmethod
    def list_of_dicts_to_dict(list_of_dict: list, key) -> dict:
        """
        Function takes a list of dicts and returns a single dict.

        Parameters:
        list_of_dict: list of dicts, i.e. [{'key1': 'value1', 'key2': 'value2'}, {'key1': 'value1', 'key2': 'value2'}]
        key (str): key to use for the dict. i.e. 'key1'
            in our example, [{'key1': 'value1', 'key2': 'value2'}, {'key1': 'value1', 'key2': 'value2'}]
            The output dictionary would have value1, value2 as the keys:
            {'value1': {'key1': 'value1', 'key2': 'value2'}, 'value2': {'key1': 'value1', 'key2': 'value2'}}

        """
        obj_dict = {}
        __empty = [obj_dict.update({x[key]: x}) for x in list_of_dict]
        return obj_dict

    @staticmethod
    def cartesian_join(pd1, pd2):
        df1 = pd1.copy()
        df2 = pd2.copy()
        df1.reset_index()
        df2.reset_index()
        df1['cartesian_join_key'] = 1
        df2['cartesian_join_key'] = 1
        result = pd.merge(df1, df2, on ='cartesian_join_key').drop(labels="cartesian_join_key", axis=1)
        return result
        
    def sub_request_params (self, request: str) -> str:
        substitutions_dict = {
            '{{AUTHORIZATION}}': self.APIheaders['authorization'],
            '{{DNS}}': self.network_DNS,
            '{{version}}': self.LatestAPIversion
        }
        for key, value in substitutions_dict.items():
            request = request.replace(key, value)
        return request
    
    def __parse_network_api_json(self, network_api: dict):
        output = {}
        for key, value in network_api.items():
            if isinstance(value, dict):
                output[key] = self.__parse_network_api_json(value)
            elif isinstance(value, list):
                first_key_set = set()
                for item in value:
                    if isinstance(item, dict):
                        first_key_set.add(list(item.keys())[0])
                if len(first_key_set) == 1:
                    output[key] = self.__parse_network_api_json(self.list_of_dicts_to_dict(value, list(first_key_set)[0]))
                else:
                    output[key] = value
            else:
                output[key] = value
        return output


from sys import platform
import requests
import pandas as pd
import os
import json
from urllib.parse import urlparse

class Vv:
    def __init__(self):
        self.vaultURL = None
        self.vaultUserName = None
        self.vaultPassword = None
        self.vaultConnection = None
        self.sessionId = None
        self.vaultId: str = None
        self.vaultDNS: str = None
        self.APIheaders = None
        self.APIversionList = []
        self.LatestAPIversion = 'v21.3'
#         self.vaultObjects = None
#         self.all_references_metadata = None
#         self.all_references_names = None
#         self.vault_references_all = None



    def authenticate(self, 
                     vaultURL=None, 
                     vaultUserName=None, 
                     vaultPassword=None, 
                     sessionId=None,
                     vaultId=None,
                     if_return=False, *args, **kwargs):
        """
        TODO: Docs
        """

        
        self.vaultURL = self.vaultURL if vaultURL is None else vaultURL
        self.vaultUserName = self.vaultUserName if vaultUserName is None else vaultUserName
        self.vaultPassword = self.vaultPassword if vaultPassword is None else vaultPassword
        self.sessionId = self.sessionId if sessionId is None else sessionId
        self.vaultId = self.vaultId if vaultId is None else vaultId
        
        url_parse = urlparse(self.vaultURL)
        if len(url_parse.scheme) == 0:
            self.network_protocol = 'https'
            if len(url_parse.path) > 0:
                self.vaultDNS = url_parse.path
                self.vaultURL = self.network_protocol + '://' + url_parse.path

        if len(url_parse.scheme) > 0:
            self.network_protocol = url_parse.scheme
            if len(url_parse.netloc) > 0:
                self.vaultDNS = url_parse.netloc
                self.vaultURL = url_parse.scheme + '://' + url_parse.netloc

        if (self.vaultURL is None) or (len(self.vaultURL) == 0):
            raise Exception('vaultURL is required')
        
        if (self.vaultUserName and self.vaultPassword and self.vaultURL):
            pload = {'username': self.vaultUserName,'password': self.vaultPassword}
            self.vaultConnection = requests.post(f'{self.vaultURL}/api/{self.LatestAPIversion}/auth',data = pload)
            self.sessionId = self.vaultConnection.json()['sessionId']
            self.vaultId = self.vaultConnection.json()['vaultId']
            
        self.APIheaders = {'Authorization': self.sessionId}
        self.APIversionList = []
        
        # Error checking whether the required parameters are passed in
        # The check happens here because this is where all the self assignments has completed
        if (not (self.vaultId and self.sessionId and self.vaultURL)) and (not (self.vaultUserName and self.vaultPassword and  self.vaultURL)):
            raise Exception("Please provide either vaultId, sessionId and vaultURL or vaultUserName, vaultPassword and vaultURL")
        
        for API in requests.get(self.vaultURL +'/api', headers=self.APIheaders).json()['values'].keys():
            self.APIversionList.append(float(API.replace("v", "")))
        self.APIversionList.sort()
        self.LatestAPIversion = "v" + str(self.APIversionList[-1])
        
        if if_return:
            return {'vaultURL':self.vaultURL, 
                    'vaultUserName':self.vaultUserName, 
                    'vaultPassword':self.vaultPassword, 
                    'vaultConnection':self.vaultConnection, 
                    'sessionId':self.sessionId, 
                    'APIheaders':self.APIheaders, 
                    'APIversionList':self.APIversionList, 
                    'LatestAPIversion':self.LatestAPIversion}
            
            
    def query(self, query):
        url = f"{self.vaultURL}/api/{self.LatestAPIversion}/query"
        
        h = {
        "X-VaultAPI-DescribeQuery":"true",
        "Content-Type":"application/x-www-form-urlencoded",
            "Accept": "application/json",
            "Authorization": self.sessionId
        }
        params = {
        "q":query
        }

        r = pd.DataFrame(requests.get(url, headers=h, params=params).json()['data'])
        
        return r
    
    def bulk_query(self, query):
        url = f"{self.vaultURL}/api/{self.LatestAPIversion}/query"
        
        h = {
        "X-VaultAPI-DescribeQuery":"true",
        "Content-Type":"application/x-www-form-urlencoded",
            "Accept": "application/json",
            "Authorization": self.sessionId
        }
        params = {
        "q":query
        }

        r = requests.get(url, headers=h, params=params).json()
        
        output = pd.DataFrame(r['data'])
        
        try:
            next_page_url = r['responseDetails']['next_page'][:-4]
            more_pages = True
            page_count = 1000
            
            while more_pages:
                r = pd.DataFrame(requests.get(f"{self.vaultURL}"+ next_page_url+ str(page_count), headers=h).json()['data'])
                if len(r) == 0:
                    more_pages = False
                else:
                    output = pd.concat([output,r],ignore_index=True).copy()
                    page_count += 1000
        except:
            pass
        
        return output
    
    def object_field_metadata(self, object_api_name):
        url = f"{self.vaultURL}/api/{self.LatestAPIversion}/metadata/vobjects/{object_api_name}"
        r = requests.get(url, headers = self.APIheaders).json()['object']['fields']
        return pd.DataFrame(r)
    
    def describe_objects(self):
        url = f"{self.vaultURL}/api/{self.LatestAPIversion}/metadata/vobjects"
        r = requests.get(url, headers = self.APIheaders).json()['objects']
        return pd.DataFrame(r).sort_values(by='name')
        


class NetworkAPIHeaderValueNotFound(Exception):
    """Exception raised when the required values are not provided to the
    Vn.api_call() method.

    Expecting to find a api['request']['header'][key]['value'] attribute in the json/dict provided, 
    where 'key' represents Authorization or Content Type header.
    """

    def __init__(self, message="The defined API does not contain a value parameter in its header.\
Expecting to find a api['request']['header'][key]['value'] attribute in the json/dict provided, \
where 'key' represents Authorization or Content Type header."):
        self.message = message
        super().__init__(self.message)

class NetworkAPIRequestMethodNotFound(Exception):
    """Exception raised when the required values are not provided to the
    Vn.api_call() method.

    Required Values:
        Expecting to find a api['request']['method'] attribute in the json/dict provided.
    """

    def __init__(self, message="The defined API does not contain a method parameter in its request method.\
Expecting to find a api['request']['method'] attribute in the json/dict provided."):
        self.message = message
        super().__init__(self.message)

class NetworkAPIRequestURLNotFound(Exception):
    """Exception raised when the required values are not provided to the
    Vn.api_call() method.

    Required Values:
        Expecting to find a api['request']['url']['raw'] attribute in the json/dict provided.
    """

    def __init__(self, message="The defined API does not contain a method parameter in its request method.\
Expecting to find a api['request']['url']['raw'] attribute in the json/dict provided."):
        self.message = message
        super().__init__(self.message)

class NetworkAPIRequestBodyNotFound(Exception):
    """Exception raised when the required values are not provided to the
    Vn.api_call() method.

    Required Values:
        The parameter body: dict is required in this API call, please refer to the Network API documentation for the syntax of the body parameter or set the if_example parameter to True for an example.
    """

    def __init__(self, message="The parameter body: dict is is required in this API call, please refer to the Network API documentation for the syntax of the body parameter or set the if_example parameter to True for an example."):
        self.message = message
        super().__init__(self.message)

class NetworkAPIRequestVariableNotFound(Exception):
    """Exception raised when the required values are not provided to the
    Vn.api_call() method.

    Required Values:
        The parameter variables: dict is required in this API call, please refer to the Network API documentation for the syntax of the body parameter or set the if_example parameter to True for an example.
    """

    def __init__(self, message="The parameter variables: dict is required in this API call, please refer to the Network API documentation for the syntax of the body parameter or set the if_example parameter to True for an example."):
        self.message = message
        super().__init__(self.message)

class NetworkAPIRequestQueryNotFound(Exception):
    """Exception raised when the required values are not provided to the
    Vn.api_call() method.

    Required Values:
        The parameter queries: dict is required in this API call, please refer to the Network API documentation for the syntax of the body parameter or set the if_example parameter to True for an example.
    """

    def __init__(self, message="The parameter queries: dict is required in this API call, please refer to the Network API documentation for the syntax of the body parameter or set the if_example parameter to True for an example."):
        self.message = message
        super().__init__(self.message)





import asyncio
from functools import wraps, partial

async def async_parallel(func, args):
    # runs the function(s) in parallel
    return await asyncio.gather(*[func(arg) for arg in args])

async def async_serial(func, args):
    # runs the function(s) in serial order, 
    # awaiting for each iteration's completion before executing the next.
    return [await func(arg) for arg in args]

def async_wrap(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)
    return run 


import pandas as pd
from collections import OrderedDict
from typing import Union, Iterable, final

# ================================================
# Account Object Functions
# ================================================

def object_account_drop_personaccount_cols(df: pd.DataFrame) -> pd.DataFrame:
    # Drops Person Account columns from the DataFrame
    person_account_list = ['FirstName', 'LastName', 'MiddleName', 'PersonDoNotCall','PersonIndividualId','PersonAssistantName', 'PersonAssistantPhone','PersonBirthdate', 'PersonBirthDate', 'PersonContactId', 'PersonDepartment', 'PersonEmail', 'PersonEmailBouncedDate', 'PersonEmailBouncedReason','PersonHasOptedOutOfFax', 'PersonHasOptedOutOfEmail', 'PersonHomePhone', 'PersonLeadSource', 'PersonMailingAddress', 'PersonMailingCity',
                           'PersonMailingGeocodeAccuracy', 'PersonMailingLatitude', 'PersonMailingLongitude', 'PersonMailingStreet', 'PersonMobilePhone', 'PersonOtherCity', 'PersonOtherCountry', 'PersonOtherPostalCode', 'PersonOtherState', 'PersonOtherCountryCode', 'PersonOtherStateCode', 'PersonOtherLatitude', 'PersonOtherLongitude', 'PersonOtherPhone', 'PersonOtherStreet', 'PersonTitle', 'Suffix']
    final_drop = []
    for column in df.columns:
        if column[-4:] == '__pc':
            final_drop.append(column)
        elif column in person_account_list:
            final_drop.append(column)
    
    return df.drop(columns=final_drop, axis=1)

def object_account_drop_businessaccount_cols(df: pd.DataFrame) -> pd.DataFrame:
    # Drops Person Account columns from the DataFrame
    business_account_list = ['DunsNumber','Jigsaw','NaicsCode','NaicsDesc','Name','Sic','Tradestyle','SicDesc','TickerSymbol']
    final_drop = []
    for column in df.columns:
        if column in business_account_list:
            final_drop.append(column)
    return df.drop(columns=final_drop, axis=1)

def unpack_column(result: pd.DataFrame, column: str='errors'):
    return pd.concat([result, result[column].apply(lambda row: pd.Series(row[0]) if len(row) >0 else pd.Series(dtype='int64')).fillna("")], axis=1).drop(columns=[column])

def metadata_parse(metadata_value: Union[pd.DataFrame, Iterable, OrderedDict, bool, list, int, None]) -> Union[Iterable, OrderedDict, dict, bool, list, int, None]:
    """
    Function recursively traverses down a set of nested DataFrame, Dict, List, Ordered Dict iterables until
    it reaches a str, bool, empty list, or None value to return a dict comprising of the original hierarchical data.
    """
    if type(metadata_value) == str:
        return metadata_value
    elif type(metadata_value) == bool:
        return metadata_value
    elif metadata_value is None:
        return None
    elif type(metadata_value) == int:
        return metadata_value
    elif type(metadata_value) == list:
        if len(metadata_value) == 0:
            return []
        else:
            working_list = []
            for item in metadata_value:
                working_list.append(metadata_parse(item))
            return working_list
    elif type(metadata_value) == OrderedDict:
        working_dict = {}
        for key in metadata_value.keys():
            working_dict[key] = metadata_parse(metadata_value[key])
        return working_dict
    elif type(metadata_value) == pd.DataFrame:
        working_df_dict = {}
        for column in metadata_value.columns:
            if len(metadata_value[column]) == 0:
                working_df_dict[column] = None
            elif len(metadata_value[column]) > 1:
                for item in metadata_value[column]:
                    working_df_dict[column] = metadata_parse(item)
            elif len(metadata_value[column]) == 1:
                working_df_dict[column] = metadata_parse(metadata_value[column][0])
        return working_df_dict



try:
    from salesforce.utilities.df_utils import *
    from salesforce.utilities.async_utils import *
    from salesforce.utilities.sf_query_processors import *
except:
    from utilities.df_utils import *
    from utilities.async_utils import *
    from utilities.sf_query_processors import *


def serialze_zeep(func):
    import zeep.helpers
    import pandas as pd
    def wrapper(*args, **kwargs):
        result = pd.DataFrame.from_dict(zeep.helpers.serialize_object(func(*args, **kwargs)),orient="index").transpose()
        return result
    return wrapper